from copy import deepcopy

import pytest
from lxml import etree

from translate.misc.multistring import multistring
from translate.storage import aresource, test_monolingual
from translate.storage.base import TranslationStore


class TestAndroidResourceUnit(test_monolingual.TestMonolingualUnit):
    UnitClass = aresource.AndroidResourceUnit

    def __check_escape(self, string, xml, target_language=None):
        """Helper that checks that a string is output with the right escape."""
        unit = self.UnitClass("teststring")

        if target_language is not None:
            store = TranslationStore()
            store.settargetlanguage(target_language)
            unit._store = store

        unit.target = string

        print("unit.target:", repr(unit.target))
        print("xml:", repr(xml))

        assert str(unit) == xml

    def __check_parse(self, string, xml):
        """Helper that checks that a string is parsed correctly."""
        parser = etree.XMLParser(strip_cdata=False)

        translatable = 'translatable="false"' not in xml
        et = etree.fromstring(xml, parser)
        unit = self.UnitClass.createfromxmlElement(et)

        print("unit.target:", repr(unit.target))
        print("string:", string)
        print("translatable:", repr(unit.istranslatable()))

        assert unit.target == string
        assert unit.istranslatable() == translatable

    ############################ Check string escape ##########################

    def test_escape_message_with_newline(self):
        string = "message\nwith newline"
        xml = '<string name="teststring">message\n\\nwith newline</string>\n'
        self.__check_escape(string, xml)

    def test_escape_quotes_with_newline(self):
        string = "'message'\nwith newline"
        xml = "<string name=\"teststring\">\\'message\\'\n\\nwith newline</string>\n"
        self.__check_escape(string, xml)

    def test_escape_message_with_newline_in_xml(self):
        string = "message\n\nwith newline in xml\n"
        xml = (
            '<string name="teststring">message\n\\n\n\\nwith newline in xml\n\\n'
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_twitter(self):
        string = "@twitterescape"
        xml = '<string name="teststring">\\@twitterescape</string>\n'
        self.__check_escape(string, xml)

    def test_escape_quote(self):
        string = "quote 'escape'"
        xml = "<string name=\"teststring\">quote \\'escape\\'</string>\n"
        self.__check_escape(string, xml)

    def test_escape_question(self):
        string = "question?"
        xml = '<string name="teststring">question\\?</string>\n'
        self.__check_escape(string, xml)

    def test_escape_double_space(self):
        string = "double  space"
        xml = '<string name="teststring">"double  space"</string>\n'
        self.__check_escape(string, xml)

    def test_escape_leading_space(self):
        string = " leading space"
        xml = '<string name="teststring">" leading space"</string>\n'
        self.__check_escape(string, xml)

    def test_escape_tailing_space(self):
        string = "tailing space "
        xml = '<string name="teststring">"tailing space "</string>\n'
        self.__check_escape(string, xml)

    def test_escape_xml_entities(self):
        string = ">xml&entities"
        xml = '<string name="teststring">&gt;xml&amp;entities</string>\n'
        self.__check_escape(string, xml)

    def test_escape_html_code(self):
        string = "some <b>html code</b> here"
        xml = '<string name="teststring">some <b>html code</b> here' "</string>\n"
        self.__check_escape(string, xml)

    def test_escape_html_code_quote(self):
        string = "some <b>html code</b> 'here'"
        xml = (
            "<string name=\"teststring\">some <b>html code</b> \\'here\\'" "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_html_code_quote_newline(self):
        string = "some \n<b>html code</b> 'here'"
        xml = (
            "<string name=\"teststring\">some \n\\n<b>html code</b> \\'here\\'"
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_arrows(self):
        string = "<<< arrow"
        xml = '<string name="teststring">&lt;&lt;&lt; arrow</string>\n'
        self.__check_escape(string, xml)

    def test_escape_link(self):
        string = '<a href="http://example.net">link</a>'
        xml = (
            '<string name="teststring">\n'
            '  <a href="http://example.net">link</a>\n'
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_link_and_text(self):
        string = '<a href="http://example.net">link</a> and text'
        xml = (
            '<string name="teststring"><a href="http://example.net">link'
            "</a> and text</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_blank_string(self):
        string = ""
        xml = '<string name="teststring"></string>\n'
        self.__check_escape(string, xml)

    def test_plural_escape_message_with_newline(self):
        mString = multistring(
            ["one message\nwith newline", "other message\nwith newline"]
        )
        xml = (
            '<plurals name="teststring">\n'
            '    <item quantity="one">one message\n\\nwith newline</item>'
            '<item quantity="other">other message\n\\nwith newline</item>'
            "</plurals>\n"
        )
        self.__check_escape(mString, xml, "en")

    def test_plural_invalid_lang(self):
        mString = multistring(["one message", "other message"])
        xml = (
            '<plurals name="teststring">\n'
            '    <item quantity="one">one message</item>'
            '<item quantity="other">other message</item>'
            "</plurals>\n"
        )
        self.__check_escape(mString, xml, "invalid")

    def test_escape_html_quote(self):
        string = "start 'here' <b>html code 'to escape'</b> also 'here'"
        xml = (
            "<string name=\"teststring\">start \\'here\\' <b>html code \\'to escape\\'</b> also \\'here\\'"
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_html_leading_space(self):
        string = " <b>html code 'to escape'</b> some 'here'"
        xml = (
            "<string name=\"teststring\"> <b>html code \\'to escape\\'</b> some \\'here\\'"
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_html_trailing_space(self):
        string = "<b>html code 'to escape'</b> some 'here' "
        xml = (
            "<string name=\"teststring\"><b>html code \\'to escape\\'</b> some \\'here\\' "
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_html_with_ampersand(self):
        string = "<b>html code 'to escape'</b> some 'here' with &amp; char"
        xml = (
            "<string name=\"teststring\"><b>html code \\'to escape\\'</b> some \\'here\\' with &amp; char"
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_html_double_space(self):
        string = "<b>html code 'to  escape'</b> some 'here'"
        xml = (
            "<string name=\"teststring\"><b>html code \\'to \\u0020escape\\'</b> some \\'here\\'"
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_html_deep_double_space(self):
        string = "<b>html code 'to  <i>escape</i>'</b> some 'here'"
        xml = (
            "<string name=\"teststring\"><b>html code \\'to \\u0020<i>escape</i>\\'</b> some \\'here\\'"
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_complex_xml(self):
        string = '<g:test xmlns:g="ttt" g:somevalue="aaaa  &quot;  aaa">value</g:test> &amp; outer &gt; <br/>text'
        xml = (
            '<string name="teststring">'
            '<g:test xmlns:g="ttt" g:somevalue="aaaa  &quot;  aaa">value</g:test> &amp; outer &gt; <br/>text'
            "</string>\n"
        )
        self.__check_escape(string, xml)

    def test_escape_quoted_newlines(self):
        self.__check_escape(
            "\n\nstring with newlines",
            r"""<string name="teststring">"
\n
\nstring with newlines"</string>
""",
        )

    ############################ Check string parse ###########################

    def test_parse_message_with_newline(self):
        string = "message\nwith newline"
        xml = '<string name="teststring">message\\nwith newline</string>\n'
        self.__check_parse(string, xml)

    def test_parse_message_with_newline_in_xml(self):
        string = "message\nwith\n newline\n in xml"
        xml = (
            '<string name="teststring">message\n\\nwith\\n\nnewline\\n\nin xml'
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_twitter(self):
        string = "@twitterescape"
        xml = '<string name="teststring">\\@twitterescape</string>\n'
        self.__check_parse(string, xml)

    def test_parse_question(self):
        string = "?"
        xml = '<string name="question">\\?</string>'
        self.__check_parse(string, xml)

    def test_parse_quote(self):
        string = "quote 'escape'"
        xml = "<string name=\"teststring\">quote \\'escape\\'</string>\n"
        self.__check_parse(string, xml)

    def test_parse_double_space(self):
        string = "double  space"
        xml = '<string name="teststring">"double  space"</string>\n'
        self.__check_parse(string, xml)

    def test_parse_leading_space(self):
        string = " leading space"
        xml = '<string name="teststring">" leading space"</string>\n'
        self.__check_parse(string, xml)

    def test_parse_quoted_newlines(self):
        self.__check_parse(
            "\n\nstring with newlines",
            r"""<string name="teststring">"
\n
\nstring with newlines"</string>
""",
        )

    def test_parse_xml_entities(self):
        string = ">xml&entities"
        xml = '<string name="teststring">&gt;xml&amp;entities</string>\n'
        self.__check_parse(string, xml)

    def test_parse_html_code(self):
        string = "some <b>html code</b> here"
        xml = '<string name="teststring">some <b>html code</b> here' "</string>\n"
        self.__check_parse(string, xml)

    def test_parse_arrows(self):
        string = "<<< arrow"
        xml = '<string name="teststring">&lt;&lt;&lt; arrow</string>\n'
        self.__check_parse(string, xml)

    def test_parse_link(self):
        string = '<a href="http://example.net">link</a>'
        xml = (
            '<string name="teststring"><a href="http://example.net">link'
            "</a></string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_link_and_text(self):
        string = '<a href="http://example.net">link</a> and text'
        xml = (
            '<string name="teststring"><a href="http://example.net">link'
            "</a> and text</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_blank_string(self):
        string = ""
        xml = '<string name="teststring"></string>\n'
        self.__check_parse(string, xml)

    def test_parse_trailing_space(self):
        string = "test"
        xml = '<string name="teststring">test </string>\n'
        self.__check_parse(string, xml)

    def test_parse_trailing_spaces(self):
        string = "test"
        xml = '<string name="teststring">test    </string>\n'
        self.__check_parse(string, xml)

    def test_parse_leading_spaces(self):
        string = "test"
        xml = '<string name="teststring">    test</string>\n'
        self.__check_parse(string, xml)

    def test_parse_trailing_newline(self):
        string = "test"
        xml = '<string name="teststring">test\n</string>\n'
        self.__check_parse(string, xml)

    def test_parse_many_quotes(self):
        string = "test"
        xml = '<string name="teststring">""""""""""test"""""""</string>\n'
        self.__check_parse(string, xml)

    def test_parse_blank_string_again(self):
        string = ""
        xml = '<string name="teststring"/>\n'
        self.__check_parse(string, xml)

    def test_parse_double_quotes_string(self):
        """Check that double quotes got removed."""
        string = "double quoted text"
        xml = '<string name="teststring">"double quoted text"</string>\n'
        self.__check_parse(string, xml)

    def test_parse_newline_in_string(self):
        """Check that newline is read as space.

        At least it seems to be what Android does.
        """
        string = "newline\nin string"
        xml = '<string name="teststring">newline\\nin string</string>\n'
        self.__check_parse(string, xml)

    def test_parse_not_translatable_string(self):
        string = "string"
        xml = '<string name="teststring" translatable="false">string' "</string>\n"
        self.__check_parse(string, xml)

    def test_plural_parse_message_with_newline(self):
        mString = multistring(
            ["one message\nwith newline", "other message\nwith newline"]
        )
        xml = (
            '<plurals name="teststring">\n'
            '    <item quantity="one">one message\\nwith newline</item>\n'
            '    <item quantity="other">other message\\nwith newline</item>\n'
            "</plurals>\n"
        )
        self.__check_parse(mString, xml)

    def test_plural_parse_message_with_comments(self):
        mString = multistring(["one message", "other message"])
        xml = (
            '<plurals name="teststring">\n'
            "    <!-- comment of one string -->\n"
            '    <item quantity="one">one message</item>\n'
            "    <!-- comment of other string -->\n"
            '    <item quantity="other">other message</item>\n'
            "</plurals>\n"
        )
        self.__check_parse(mString, xml)

    def test_parse_html_quote(self):
        string = "start 'here' <b>html code 'to escape'</b> also 'here'"
        xml = (
            "<string name=\"teststring\">start \\'here\\' <b>html code \\'to escape\\'</b> also \\'here\\'"
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_html_leading_space(self):
        string = " <b>html code 'to escape'</b> some 'here'"
        xml = (
            "<string name=\"teststring\"> <b>html code \\'to escape\\'</b> some \\'here\\'"
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_html_leading_space_quoted(self):
        string = " <b>html code 'to escape'</b> some 'here'"
        xml = (
            '<string name="teststring">" "<b>"html code \'to escape\'"</b>" some \'here\'"'
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_html_trailing_space(self):
        string = "<b>html code 'to escape'</b> some 'here' "
        xml = (
            "<string name=\"teststring\"><b>html code \\'to escape\\'</b> some \\'here\\' "
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_html_trailing_space_quoted(self):
        string = "<b>html code 'to escape'</b> some 'here' "
        xml = (
            '<string name="teststring"><b>"html code \'to escape\'"</b>" some \'here\' "'
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_html_with_ampersand(self):
        string = "<b>html code 'to escape'</b> some 'here' with &amp; char"
        xml = (
            "<string name=\"teststring\"><b>html code \\'to escape\\'</b> some \\'here\\' with &amp; char"
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_html_double_space_quoted(self):
        string = "<b>html code 'to  escape'</b> some 'here'"
        xml = (
            "<string name=\"teststring\"><b>html code 'to \\u0020escape'</b> some 'here'"
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_html_deep_double_space_quoted(self):
        string = "<b>html code 'to  <i>  escape</i>'</b> some 'here'"
        xml = (
            '<string name="teststring"><b>"html code \'to  "<i>"  escape"</i>\\\'</b> some \\\'here\\\''
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_complex_xml(self):
        string = '<g:test xmlns:g="ttt" g:somevalue="aaaa  &quot;  aaa">value</g:test> outer &amp; text'
        xml = (
            '<string name="teststring">'
            '<g:test xmlns:g="ttt" g:somevalue="aaaa  &quot;  aaa">value</g:test> outer &amp; text'
            "</string>\n"
        )
        self.__check_parse(string, xml)

    def test_parse_unicode(self):
        with pytest.raises(ValueError):
            self.__check_parse("", r'<string name="test">\utest</string>')
        self.__check_parse("\u0230", r'<string name="test">\u0230</string>')


class TestAndroidResourceFile(test_monolingual.TestMonolingualStore):
    StoreClass = aresource.AndroidResourceFile

    def test_targetlanguage_default_handlings(self):
        store = self.StoreClass()

        # Initial value is None
        assert store.gettargetlanguage() is None

        # sourcelanguage shouldn't change the targetlanguage
        store.setsourcelanguage("en")
        assert store.gettargetlanguage() is None

        # targetlanguage setter works correctly
        store.settargetlanguage("de")
        assert store.gettargetlanguage() == "de"

        # explicit targetlanguage wins over filename
        store.filename = "dommy/values-it/res.xml"
        assert store.gettargetlanguage() == "de"

    def test_targetlanguage_auto_detection_filename(self):
        store = self.StoreClass()

        # Check language auto_detection
        store.filename = "project/values-it/res.xml"
        assert store.gettargetlanguage() == "it"

    def test_targetlanguage_auto_detection_filename_default_language(self):
        store = self.StoreClass()

        store.setsourcelanguage("en")

        # Check language auto_detection
        store.filename = "project/values/res.xml"
        assert store.gettargetlanguage() == "en"

    def test_targetlanguage_auto_detection_invalid_filename(self):
        store = self.StoreClass()

        store.setsourcelanguage("en")

        store.filename = "project/invalid_directory/res.xml"
        assert store.gettargetlanguage() is None

        store.filename = "invalid_directory"
        assert store.gettargetlanguage() is None

    def test_namespaces(self):
        content = """<resources xmlns:tools="http://schemas.android.com/tools">
            <string name="string1" tools:ignore="PluralsCandidate">string1</string>
            <string name="string2">string2</string>
        </resources>"""
        store = self.StoreClass()
        store.parse(content)
        newstore = self.StoreClass()
        newstore.addunit(store.units[0], new=True)
        print(newstore)
        assert b'<resources xmlns:tools="http://schemas.android.com/tools">' in bytes(
            newstore
        )

    def test_serialize(self):
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="test">Test</string>
</resources>"""
        store = self.StoreClass()
        store.parse(content)
        assert bytes(store) == content

    def test_add_formatting(self):
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="test">Test</string>
</resources>"""
        other = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="other">Test</string>
    <string name="other2">Test</string>
</resources>"""
        expected = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="test">Test</string>
    <string name="other">Test</string>
</resources>"""
        store = self.StoreClass()
        store.parse(content)
        otherstore = self.StoreClass()
        otherstore.parse(other)
        store.addunit(otherstore.units[0], True)
        assert bytes(store) == expected

    def test_entity(self):
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE resources [
<!ENTITY appName "Linphone">
]>
<resources>
    <string name="app_name">&appName;</string>
    <string name="app_core">&appName; Core</string>
</resources>"""
        store = self.StoreClass()
        store.parse(content)
        assert store.units[0].source == "&appName;"
        store.units[0].target = "&appName;"
        assert store.units[1].source == "&appName; Core"
        store.units[1].target = "&appName; Core"
        assert bytes(store) == content

    def test_invalid_entity(self):
        content = """<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE resources [
<!ENTITY appName "Linphone">
]>
<resources>
    <string name="app_name">&appName;</string>
    <string name="app_core">&appName; Core</string>
</resources>"""
        store = self.StoreClass()
        store.parse(content.encode())
        store.units[0].target = "&appName;"
        store.units[1].target = "&otherName; Core"
        assert bytes(store).decode() == content.replace(
            "&appName; Core", "&amp;otherName; Core"
        )

    def test_indent(self):
        content = """<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE resources [
<!ENTITY url_privacy_policy "http://example.com/">
]>
<resources>
    <string name="privacy_policy"><u><a href="&url_privacy_policy;">Datenschutzerklärung</a></u></string>
</resources>""".encode()
        store = self.StoreClass()
        store.parse(content)
        assert bytes(store) == content

    def test_edit_plural_markup(self):
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> den</item>
        <item quantity="few"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dny</item>
        <item quantity="other"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dnu</item>
    </plurals>
</resources>"""
        store = self.StoreClass()
        store.targetlanguage = "cs"
        store.parse(content)
        store.units[0].target = multistring(
            [
                '<xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> den',
                '<xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dny',
                '<xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dnu',
            ]
        )
        assert bytes(store) == content

    def test_entity_add(self, edit=True):
        content = """<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE resources [
<!ENTITY appName "Zkouška">
]>
<resources>
    <string name="app_name">&appName;</string>
    <string name="app_core">&appName; Core</string>
</resources>"""
        store = self.StoreClass()
        store.parse(content.encode())
        second = self.StoreClass()
        newunit = deepcopy(store.units[0])
        second.addunit(newunit, True)
        if edit:
            newunit.target = "&appName;"
        newunit = deepcopy(store.units[1])
        second.addunit(newunit, True)
        if edit:
            newunit.target = "&appName; Core"
        # The original store should be unchanged
        assert bytes(store).decode() == content
        # The new store should have same content
        assert bytes(second).decode() == content

    def test_entity_add_noedit(self):
        self.test_entity_add(edit=False)

    def test_markup_remove(self):
        template = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="privacy_policy"><u>Datenschutzerklärung</u></string>
</resources>"""
        content = template.encode("utf-8")
        newcontent = template.replace("<u>", "").replace("</u>", "").encode("utf-8")
        store = self.StoreClass()
        store.parse(content)
        assert bytes(store) == content
        store.units[0].target = "Datenschutzerklärung"
        assert bytes(store) == newcontent

    def test_markup_set(self):
        template = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <!--Multimedia tab-->
    <string name="id">Multimedia tab</string>
</resources>"""
        content = template.encode("utf-8")
        newcontent = template.replace(">Multimedia tab<", ">Other <b>tab</b><").encode(
            "utf-8"
        )
        store = self.StoreClass()
        store.parse(content)
        assert bytes(store) == content
        store.units[0].target = "Other <b>tab</b>"
        assert bytes(store) == newcontent

    def test_edit_plural_others(self):
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="teststring">
        <item quantity="one"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> den</item>
        <item quantity="few"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dny</item>
        <item quantity="many"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dnu</item>
        <item quantity="other"><xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dnu</item>
    </plurals>
</resources>"""
        store = self.StoreClass()
        store.targetlanguage = "ru"
        store.parse(content)
        store.units[0].target = multistring(
            [
                '<xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> den',
                '<xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dny',
                '<xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2" id="count">%d</xliff:g> dnu',
            ]
        )
        assert bytes(store) == content

    def test_markup_quotes_set(self):
        template = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Test</string>
</resources>"""
        content = template.encode()
        newcontent = template.replace(
            ">Test<", ">Test <b>string</b> with \\u0020space<"
        )
        store = self.StoreClass()
        store.parse(content)
        assert bytes(store) == content
        store.units[0].target = "Test <b>string</b> with  space"
        assert bytes(store).decode() == newcontent

    def test_xliff_g(self):
        template = """<?xml version="1.0" encoding="utf-8"?>
<resources>
</resources>"""
        original = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Test: <xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2">%s</xliff:g></string>
</resources>"""
        expected = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Other: <xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2">%s</xliff:g></string>
</resources>"""
        origstore = self.StoreClass()
        origstore.parse(original.encode())
        store = self.StoreClass()
        store.parse(template.encode())
        store.addunit(origstore.units[0], new=True)
        assert bytes(store).decode() == original
        store.units[
            0
        ].target = """Other: <xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2">%s</xliff:g>"""
        assert bytes(store).decode() == expected

    def test_xliff_namespace(self):
        original = """<?xml version="1.0" encoding="utf-8"?>
<resources xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2">
    <string name="id">Test</string>
</resources>"""
        expected = """<?xml version="1.0" encoding="utf-8"?>
<resources xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2">
    <string name="id">Test: <xliff:g>%s</xliff:g></string>
</resources>"""
        store = self.StoreClass()
        store.parse(original.encode())
        assert bytes(store).decode() == original
        store.units[
            0
        ].target = """Test: <xliff:g xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2">%s</xliff:g>"""
        # The namespace should be flattened as it is defined in the toplevel element
        assert bytes(store).decode() == expected

    def test_edit_plural_zh_hk(self):
        content = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="vms_num_visitors">
        <item quantity="one">%d visitor</item>
        <item quantity="other">%d visitors</item>
    </plurals>
</resources>"""
        store = self.StoreClass()
        store.targetlanguage = "zh-rHK"
        store.parse(content.encode())
        store.units[0].target = "%d 訪客"
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="vms_num_visitors">
        <item quantity="other">%d 訪客</item>
    </plurals>
</resources>"""
        )

    def test_edit_plural_b_zh_hk(self):
        content = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="vms_num_visitors">
        <item quantity="one">%d visitor</item>
        <item quantity="other">%d visitors</item>
    </plurals>
</resources>"""
        store = self.StoreClass()
        store.targetlanguage = "b+zh+Hant+HK"
        store.parse(content.encode())
        store.units[0].target = "%d 訪客"
        assert (
            bytes(store).decode()
            == """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="vms_num_visitors">
        <item quantity="other">%d 訪客</item>
    </plurals>
</resources>"""
        )

    def test_missing_plural(self):
        content = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <plurals name="vms_num_visitors">
        <item quantity="one">%d visitor</item>
        <item quantity="other">%d visitors</item>
    </plurals>
</resources>"""
        store = self.StoreClass()
        store.targetlanguage = "fr"
        store.parse(content.encode())
        assert store.units[0].target == multistring(["%d visitor", "", "%d visitors"])



#
# Copyright 2007,2009,2011 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""This module represents the Arabic language.

.. seealso:: :wp:`Arabic_language`
"""


import re

from translate.lang import common


def reverse_quotes(text):
    def convertquotation(match):
        return "”%s“" % match.group(1)

    return re.sub("“([^”]+)”", convertquotation, text)


class ar(common.Common):
    """This class represents Arabic."""

    listseperator = "، "

    puncdict = {
        ",": "،",
        ";": "؛",
        "?": "؟",
        # This causes problems with variables, so commented out for now:
        # "%": "٪",
    }

    numbertuple = (
        ("0", "٠"),  # U+0660 Arabic-Indic digit zero.
        ("1", "١"),  # U+0661 Arabic-Indic digit one.
        ("2", "٢"),  # U+0662 Arabic-Indic digit two.
        ("3", "٣"),  # U+0663 Arabic-Indic digit three.
        ("4", "٤"),  # U+0664 Arabic-Indic digit four.
        ("5", "٥"),  # U+0665 Arabic-Indic digit five.
        ("6", "٦"),  # U+0666 Arabic-Indic digit six.
        ("7", "٧"),  # U+0667 Arabic-Indic digit seven.
        ("8", "٨"),  # U+0668 Arabic-Indic digit eight.
        ("9", "٩"),  # U+0669 Arabic-Indic digit nine.
    )

    ignoretests = {
        "all": ["acronyms", "simplecaps", "startcaps"],
    }

    @classmethod
    def punctranslate(cls, text):
        text = super().punctranslate(text)
        return reverse_quotes(text)



#
# Copyright 2004-2006 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Convert Qt Linguist (.ts) files to Gettext PO localization files.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/ts2po.html
for examples and usage instructions.
"""

from translate.storage import po, ts2


class ts2po:
    def __init__(self, duplicatestyle="msgctxt", pot=False):
        self.duplicatestyle = duplicatestyle
        self.pot = pot

    def convertmessage(
        self,
        contextname,
        messagenum,
        source,
        target,
        disambiguation,
        msgcomments,
        transtype,
    ):
        """makes a pounit from the given message"""
        thepo = po.pounit(encoding="UTF-8")
        thepo.addlocation("%s#%d" % (contextname, messagenum))
        thepo.source = source
        if not self.pot:
            thepo.target = target
        if len(disambiguation) > 0:
            thepo.setcontext(disambiguation)
        if len(msgcomments) > 0:
            thepo.addnote(msgcomments)
        if transtype == "unfinished" and thepo.istranslated():
            thepo.markfuzzy()
        if transtype == "obsolete":
            # This should use the Gettext obsolete method but it would require quite a bit of work
            thepo.addnote("(obsolete)", origin="developer")
            # using the fact that -- quote -- "(this is nonsense)"
        return thepo

    def convertfile(self, inputfile):
        """converts a .ts file to .po format"""
        tsfile = ts2.tsfile(inputfile)
        thetargetfile = po.pofile()

        previouscontext = ""
        for inputunit in tsfile.units:
            contexts = inputunit.getcontext().split("\n")

            context = contexts[0]
            if context != previouscontext:
                previouscontext = context
                messagenum = 0

            disambiguation = ""
            if len(contexts) > 1:
                disambiguation = contexts[1]
            messagenum += 1
            thepo = self.convertmessage(
                context,
                messagenum,
                inputunit.source,
                inputunit.target,
                disambiguation,
                inputunit.getnotes(),
                inputunit._gettype(),
            )
            thetargetfile.addunit(thepo)

        thetargetfile.removeduplicates(self.duplicatestyle)
        return thetargetfile


def convertts(inputfile, outputfile, templates, pot=False, duplicatestyle="msgctxt"):
    """reads in stdin using fromfileclass, converts using convertorclass, writes to stdout"""
    convertor = ts2po(duplicatestyle=duplicatestyle, pot=pot)
    outputstore = convertor.convertfile(inputfile)
    if outputstore.isempty():
        return 0
    outputstore.serialize(outputfile)
    return 1


def main(argv=None):
    from translate.convert import convert

    formats = {"ts": ("po", convertts)}
    parser = convert.ConvertOptionParser(formats, usepots=True, description=__doc__)
    parser.add_duplicates_option()
    parser.passthrough.append("pot")
    parser.run(argv)


if __name__ == "__main__":
    main()



#
# Copyright 2002-2006 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Convert Gettext PO localization files to iCalendar files.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/ical2po.html
for examples and usage instructions.
"""

from translate.convert import convert
from translate.storage import ical, po


class po2ical:
    """Convert a PO file and a template iCalendar file to a iCalendar file."""

    SourceStoreClass = po.pofile
    TargetStoreClass = ical.icalfile
    TargetUnitClass = ical.icalunit
    MissingTemplateMessage = "A template iCalendar file must be provided."

    def __init__(
        self,
        input_file,
        output_file,
        template_file=None,
        include_fuzzy=False,
        output_threshold=None,
    ):
        """Initialize the converter."""
        if template_file is None:
            raise ValueError(self.MissingTemplateMessage)

        self.source_store = self.SourceStoreClass(input_file)

        self.should_output_store = convert.should_output_store(
            self.source_store, output_threshold
        )
        if self.should_output_store:
            self.include_fuzzy = include_fuzzy

            self.output_file = output_file
            self.template_store = self.TargetStoreClass(template_file)
            self.output_store = self.TargetStoreClass()

            self.source_store.makeindex()

    def merge_stores(self):
        """Convert a source file to a target file using a template file.

        Source file is in source format, while target and template files use
        target format.
        """
        for template_unit in self.template_store.units:
            for location in template_unit.getlocations():
                if location in self.source_store.locationindex:
                    source_unit = self.source_store.locationindex[location]
                    if source_unit.isfuzzy() and not self.include_fuzzy:
                        template_unit.target = template_unit.source
                    else:
                        template_unit.target = source_unit.target
                else:
                    template_unit.target = template_unit.source

    def run(self):
        """Run the converter."""
        if not self.should_output_store:
            return 0

        self.merge_stores()
        self.template_store.serialize(self.output_file)
        return 1


def run_converter(
    inputfile, outputfile, templatefile=None, includefuzzy=False, outputthreshold=None
):
    """Wrapper around converter."""
    return po2ical(
        inputfile, outputfile, templatefile, includefuzzy, outputthreshold
    ).run()


formats = {("po", "ics"): ("ics", run_converter)}


def main(argv=None):
    parser = convert.ConvertOptionParser(
        formats, usetemplates=True, description=__doc__
    )
    parser.add_threshold_option()
    parser.add_fuzzy_option()
    parser.run(argv)


if __name__ == "__main__":
    main()



from translate.search import match
from translate.storage import csvl10n


class TestMatch:
    """Test the matching class"""

    @staticmethod
    def candidatestrings(units):
        """returns only the candidate strings out of the list with (score, string) tuples"""
        return [unit.source for unit in units]

    @staticmethod
    def buildcsv(sources, targets=None):
        """Build a csvfile store with the given source and target strings"""
        if targets is None:
            targets = sources
        else:
            assert len(sources) == len(targets)
        csvfile = csvl10n.csvfile()
        for source, target in zip(sources, targets):
            unit = csvfile.addsourceunit(source)
            unit.target = target
        return csvfile

    def test_matching(self):
        """Test basic matching"""
        csvfile = self.buildcsv(["hand", "asdf", "fdas", "haas", "pond"])
        matcher = match.matcher(csvfile)
        candidates = self.candidatestrings(matcher.matches("hond"))
        candidates.sort()
        assert candidates == ["hand", "pond"]
        message = "Ek skop die bal"
        csvfile = self.buildcsv(
            [
                "Hy skop die bal",
                message,
                "Jannie skop die bal",
                "Ek skop die balle",
                "Niemand skop die bal nie",
            ]
        )
        matcher = match.matcher(csvfile)
        candidates = self.candidatestrings(matcher.matches(message))
        assert len(candidates) == 3
        # test that the 100% match is indeed first:
        assert candidates[0] == message
        candidates.sort()
        assert candidates[1:] == ["Ek skop die balle", "Hy skop die bal"]

    def test_multiple_store(self):
        """Test using multiple datastores"""
        csvfile1 = self.buildcsv(["hand", "asdf", "fdas"])
        csvfile2 = self.buildcsv(["haas", "pond"])
        matcher = match.matcher([csvfile1, csvfile2])
        candidates = self.candidatestrings(matcher.matches("hond"))
        candidates.sort()
        assert candidates == ["hand", "pond"]
        message = "Ek skop die bal"
        csvfile1 = self.buildcsv(["Hy skop die bal", message, "Jannie skop die bal"])
        csvfile2 = self.buildcsv(["Ek skop die balle", "Niemand skop die bal nie"])
        matcher = match.matcher([csvfile1, csvfile2])
        candidates = self.candidatestrings(matcher.matches(message))
        assert len(candidates) == 3
        # test that the 100% match is indeed first:
        assert candidates[0] == message
        candidates.sort()
        assert candidates[1:] == ["Ek skop die balle", "Hy skop die bal"]

    def test_extendtm(self):
        """Test that we can extend the TM after creation."""
        message = "Open file..."
        csvfile1 = self.buildcsv(["Close application", "Do something"])
        matcher = match.matcher([csvfile1])
        candidates = self.candidatestrings(matcher.matches(message))
        assert len(candidates) == 0
        csvfile2 = self.buildcsv(["Open file"])
        matcher.extendtm(csvfile2.units, store=csvfile2)
        candidates = self.candidatestrings(matcher.matches(message))
        assert len(candidates) == 1
        assert candidates[0] == "Open file"

    def test_terminology(self):
        csvfile = self.buildcsv(["file", "computer", "directory"])
        matcher = match.terminologymatcher(csvfile)
        candidates = self.candidatestrings(
            matcher.matches("Copy the files from your computer")
        )
        candidates.sort()
        assert candidates == ["computer", "file"]

    def test_brackets(self):
        """Tests that brackets at the end of a term are ignored"""
        csvfile = self.buildcsv(["file (noun)", "ISP (Internet Service Provider)"])
        matcher = match.terminologymatcher(csvfile)
        candidates = self.candidatestrings(matcher.matches("Open File"))
        assert candidates == ["file"]
        candidates = self.candidatestrings(matcher.matches("Contact your ISP"))
        # we lowercase everything - that is why we get it back differently.
        # we don't change the target text, though
        assert candidates == ["isp"]

    def test_past_tences(self):
        """Tests matching of some past tenses"""
        csvfile = self.buildcsv(["submit", "certify"])
        matcher = match.terminologymatcher(csvfile)
        candidates = self.candidatestrings(matcher.matches("The bug was submitted"))
        assert candidates == ["submit"]
        candidates = self.candidatestrings(matcher.matches("The site is certified"))

    def test_space_mismatch(self):
        """Tests that we can match with some spacing mismatch"""
        csvfile = self.buildcsv(["down time"])
        matcher = match.terminologymatcher(csvfile)
        candidates = self.candidatestrings(matcher.matches("%d minutes downtime"))
        assert candidates == ["downtime"]

    def test_hyphen_mismatch(self):
        """Tests that we can match with some spacing mismatch"""
        csvfile = self.buildcsv(["pre-order"])
        matcher = match.terminologymatcher(csvfile)
        candidates = self.candidatestrings(matcher.matches("You can preorder"))
        assert candidates == ["preorder"]
        candidates = self.candidatestrings(matcher.matches("You can pre order"))
        assert candidates == ["pre order"]

        csvfile = self.buildcsv(["pre order"])
        matcher = match.terminologymatcher(csvfile)
        candidates = self.candidatestrings(matcher.matches("You can preorder"))
        assert candidates == ["preorder"]
        candidates = self.candidatestrings(matcher.matches("You can pre order"))
        assert candidates == ["pre order"]



#
# Copyright 2021 Jack Grigg
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

r"""Manage the Fluent translation format.

It is a monolingual base class derived format with :class:`FluentFile`
and :class:`FluentUnit` providing file and unit level access.
"""

from fluent.syntax import FluentParser, ast, parse, serialize, visitor
from fluent.syntax.serializer import serialize_pattern, serialize_placeable
from fluent.syntax.stream import FluentParserStream
from translate.storage import base


def id_from_source(source):
    # If the caller does not provide a unit ID, we need to generate one
    # ourselves. A valid Fluent identifier has the following EBNF grammar:
    #     Identifier          ::= [a-zA-Z] [a-zA-Z0-9_-]*
    #
    # This means we can't simply use the source string itself as the identifier
    # (as e.g. PO files do). Instead, we hash the source string with a
    # collision-resistant hash function.
    import hashlib

    return "gen-" + hashlib.sha256(source.encode()).hexdigest()


def source_from_entry(entry):
    # Serialized patterns come in two forms:
    # - Single-line patterns, which have a leading space we need to strip (for
    #   consistency with the expectations of what callers will set).
    # - Multiline patterns, which have a leading newline we need to preserve.
    return serialize_pattern(entry.value).lstrip(" ")


class FluentUnit(base.TranslationUnit):
    """A Fluent message."""

    def __init__(self, source=None, entry=None):
        super().__init__(source)
        self._type = None
        self._id = None
        self._errors = {}
        self._attributes = {}
        self._placeables = []
        if source is not None:
            self._type = ast.Message
            self._set_value(source)
            # Default to source string
            self._id = id_from_source(self.source)
        if entry is not None:
            self.parse(entry)

    def _set_value(self, value):
        self.source = value
        # The source and target are equivalent for monolingual units.
        self.target = self.source

    def getid(self):
        return self._id

    def setid(self, value):
        self._id = value

    def adderror(self, errorname, errortext):
        self._errors[errorname] = errortext

    def geterrors(self):
        return self._errors

    def isheader(self):
        return self._type in [
            ast.Comment,
            ast.GroupComment,
            ast.ResourceComment,
        ]

    def getattributes(self):
        return self._attributes

    def getplaceables(self):
        return self._placeables

    def parse(self, entry):
        this = self

        class Parser(visitor.Visitor):
            _found_id = False

            @staticmethod
            def visit_Attribute(node):
                this._attributes[node.id.name] = source_from_entry(node)

            @staticmethod
            def visit_Placeable(node):
                this._placeables.append(serialize_placeable(node))

            @staticmethod
            def visit_Comment(node):
                if this._type not in [ast.Message, ast.Term]:
                    this._type = ast.Comment
                this.addnote(node.content)

            @staticmethod
            def visit_GroupComment(node):
                this._type = ast.GroupComment
                this.addnote(node.content)

            @staticmethod
            def visit_ResourceComment(node):
                this._type = ast.ResourceComment
                this.addnote(node.content)

            def visit_Identifier(self, node):
                if not self._found_id:
                    # Only save the first identifier we encounter (the entry's
                    # value will also contain identifiers if it has selectors).
                    this._id = node.name
                    self._found_id = True

            def visit_Message(self, node):
                this._type = ast.Message
                this._set_value(source_from_entry(node))
                self.generic_visit(node)

            def visit_Term(self, node):
                this._type = ast.Term
                this._set_value(source_from_entry(node))
                self.generic_visit(node)

        Parser().visit(entry)

    def to_entry(self):
        fp = FluentParser(False)

        # Handle standalone comments separately; they don't have any values or
        # attributes, just comment text.
        if self._type in [ast.ResourceComment, ast.GroupComment, ast.Comment]:
            return (self._type)(self.getnotes())

        assert self.source is not None
        value = fp.maybe_get_pattern(FluentParserStream(self.source))

        attributes = [
            ast.Attribute(
                ast.Identifier(id),
                fp.maybe_get_pattern(FluentParserStream(value)),
            )
            for (id, value) in self._attributes.items()
        ]

        comment = None
        if self.getnotes():
            comment = ast.Comment(self.getnotes())

        return (self._type if self._type is not None else ast.Message)(
            ast.Identifier(self.getid()),
            value=value,
            attributes=attributes,
            comment=comment,
        )


class FluentFile(base.TranslationStore):
    """A Fluent file."""

    Name = "Fluent file"
    Mimetypes = []
    Extensions = ["ftl"]
    UnitClass = FluentUnit

    def __init__(self, inputfile=None, **kwargs):
        super().__init__(**kwargs)
        self.filename = getattr(inputfile, "name", "")
        if inputfile is not None:
            fluentsrc = inputfile.read()
            self.parse(fluentsrc)

    def parse(self, fluentsrc):
        resource = parse(fluentsrc.decode("utf-8"))
        for entry in resource.body:
            # Handle this unit separately if it is invalid.
            if isinstance(entry, ast.Junk):
                for annotation in entry.annotations:
                    raise ValueError(
                        f"{annotation.code}: {annotation.message} [offset {annotation.span.start}]"
                    )

            self.addunit(FluentUnit(entry=entry))

    def serialize(self, out):
        body = [unit.to_entry() for unit in self.units]
        out.write(serialize(ast.Resource(body)).encode(self.encoding))



#
# Copyright 2008-2012 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Tools and API for translation and localization engineering.

The `Translate Toolkit <http://toolkit.translatehouse.org/>`_ is created by
localizers for localizers. It contains several utilities, as well as an API for
building localization tools.

Some of the **tools** include:

- File format `converters
  <http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/index.html#converters>`_
- `Quality checking
  <http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/index.html#quality-assurance>`_
  tools
- Tools for counting, grepping, terminology extraction, and pseudo-localization

Apart from the code for the tools above, some features of the **API** include:

- Support for multiple `file formats
  <http://docs.translatehouse.org/projects/translate-toolkit/en/latest/formats/index.html>`_
- Language information and language support code (including language detection)
- Code for translation memory, terminology matching and indexed search
- Several helper classes and functions for tools built on the Translate
  Toolkit.

"""



#
# Copyright 2015 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""This module represents the Assamese language.

.. seealso:: :wp:`Assamese_language`
"""


from translate.lang import common


class code_as(common.Common):
    """This class represents Assamese."""

    numbertuple = (
        ("0", "০"),  # U+09E6 Bengali digit zero.
        ("1", "১"),  # U+09E7 Bengali digit one.
        ("2", "২"),  # U+09E8 Bengali digit two.
        ("3", "৩"),  # U+09E9 Bengali digit three.
        ("4", "৪"),  # U+09EA Bengali digit four.
        ("5", "৫"),  # U+09EB Bengali digit five.
        ("6", "৬"),  # U+09EC Bengali digit six.
        ("7", "৭"),  # U+09ED Bengali digit seven.
        ("8", "৮"),  # U+09EE Bengali digit eight.
        ("9", "৯"),  # U+09EF Bengali digit nine.
    )



import warnings
from io import BytesIO

from translate.storage import oo


def test_makekey():
    """checks the makekey function for consistency"""
    assert (
        oo.makekey(
            (
                "project",
                r"path\to\the\sourcefile.src",
                "resourcetype",
                "GROUP_ID",
                "LOCAL_ID",
                "platform",
            ),
            False,
        )
        == "sourcefile.src#GROUP_ID.LOCAL_ID.resourcetype"
    )
    # Testwith long_key i.e. used in multifile options
    assert (
        oo.makekey(
            (
                "project",
                r"path\to\the\sourcefile.src",
                "resourcetype",
                "GROUP_ID",
                "LOCAL_ID",
                "platform",
            ),
            True,
        )
        == "project/path/to/the/sourcefile.src#GROUP_ID.LOCAL_ID.resourcetype"
    )
    assert (
        oo.makekey(
            (
                "project",
                r"path\to\the\sourcefile.src",
                "resourcetype",
                "GROUP_ID",
                "",
                "platform",
            ),
            False,
        )
        == "sourcefile.src#GROUP_ID.resourcetype"
    )
    assert (
        oo.makekey(
            (
                "project",
                r"path\to\the\sourcefile.src",
                "resourcetype",
                "",
                "LOCAL_ID",
                "platform",
            ),
            False,
        )
        == "sourcefile.src#LOCAL_ID.resourcetype"
    )
    assert (
        oo.makekey(
            (
                "project",
                r"path\to\the\sourcefile.src",
                "",
                "GROUP_ID",
                "LOCAL_ID",
                "platform",
            ),
            False,
        )
        == "sourcefile.src#GROUP_ID.LOCAL_ID"
    )
    assert (
        oo.makekey(
            ("project", r"path\to\the\sourcefile.src", "", "GROUP_ID", "", "platform"),
            False,
        )
        == "sourcefile.src#GROUP_ID"
    )


def test_escape_help_text():
    """Check the help text escape function"""
    assert (
        oo.escape_help_text("If we don't know <tag> we don't <br> escape it")
        == "If we don't know <tag> we don't <br> escape it"
    )
    # Bug 694
    assert oo.escape_help_text("A szó: <nyelv>") == "A szó: <nyelv>"
    assert (
        oo.escape_help_text(
            """...következő: "<kiszolgáló> <témakör> <elem>", ahol..."""
        )
        == """...következő: "<kiszolgáló> <témakör> <elem>", ahol..."""
    )
    # See bug 694 comments 8-10 not fully resolved.
    assert (
        oo.escape_help_text(r"...törtjel (\) létrehozásához...")
        == r"...törtjel (\\) létrehozásához..."
    )


class TestOO:
    @staticmethod
    def ooparse(oosource):
        """helper that parses oo source without requiring files"""
        dummyfile = BytesIO(oosource.encode())
        return oo.oofile(dummyfile)

    def ooregen(self, oosource):
        """helper that converts oo source to oofile object and back"""
        oofile = self.ooparse(oosource)
        return bytes(oofile).decode("utf-8")

    def test_simpleentry(self):
        """checks that a simple oo entry is parsed correctly"""
        oosource = r"svx	source\dialog\numpages.src	0	string	RID_SVXPAGE_NUM_OPTIONS	STR_BULLET			0	en-US	Character				20050924 09:13:58"
        oofile = self.ooparse(oosource)
        assert len(oofile.units) == 1
        oe = oofile.units[0]
        assert list(oe.languages.keys()) == ["en-US"]
        ol = oofile.oolines[0]
        assert ol.getkey() == (
            "svx",
            r"source\dialog\numpages.src",
            "string",
            "RID_SVXPAGE_NUM_OPTIONS",
            "STR_BULLET",
            "",
        )
        assert ol.text == "Character"
        assert str(ol) == oosource

    def test_simpleentry_quickhelptest(self):
        """checks that a simple entry with quickhelptext is parsed correctly"""
        oosource = r"sd	source\ui\dlg\sdobjpal.src	0	imagebutton	FLTWIN_SDOBJPALETTE	BTN_SYMSIZE			16	en-US	-		Toggle Symbol Size		20051017 21:40:56"
        oofile = self.ooparse(oosource)
        assert len(oofile.units) == 1
        oe = oofile.units[0]
        assert list(oe.languages.keys()) == ["en-US"]
        ol = oofile.oolines[0]
        assert ol.getkey() == (
            "sd",
            r"source\ui\dlg\sdobjpal.src",
            "imagebutton",
            "FLTWIN_SDOBJPALETTE",
            "BTN_SYMSIZE",
            "",
        )
        assert ol.quickhelptext == "Toggle Symbol Size"
        assert str(ol) == oosource

    def test_simpleentry_title(self):
        """checks that a simple entry with title text is parsed correctly"""
        oosource = r"dbaccess	source\ui\dlg\indexdialog.src	0	querybox	QUERY_SAVE_CURRENT_INDEX				0	en-US	Do you want to save the changes made to the current index?			Exit Index Design	20051017 21:40:56"
        oofile = self.ooparse(oosource)
        assert len(oofile.units) == 1
        oe = oofile.units[0]
        assert list(oe.languages.keys()) == ["en-US"]
        ol = oofile.oolines[0]
        assert ol.getkey() == (
            "dbaccess",
            r"source\ui\dlg\indexdialog.src",
            "querybox",
            "QUERY_SAVE_CURRENT_INDEX",
            "",
            "",
        )
        assert ol.title == "Exit Index Design"
        assert str(ol) == oosource

    def test_blankline(self):
        """checks that a blank line is parsed correctly"""
        oosource = "\n"
        warnings.simplefilter("error")
        oofile = self.ooparse(oosource)
        assert len(oofile.units) == 0

    def test_fieldlength(self):
        """checks that we process the length field correctly"""
        # Since the actual field is 18 characters long and the field width in this example is 16 we're not sure if they even use this!
        oosource = r"sd	source\ui\dlg\sdobjpal.src	0	imagebutton	FLTWIN_SDOBJPALETTE	BTN_SYMSIZE			16	en-US	-		Toggle Symbol Size		20051017 21:40:56"
        oofile = self.ooparse(oosource)
        assert len(oofile.units) == 1
        oe = oofile.units[0]
        assert list(oe.languages.keys()) == ["en-US"]
        ol = oofile.oolines[0]
        assert int(ol.width) == 16

    def test_escapes(self):
        """checks that we escape properly"""
        oosource = r"svx	source\dialog\numpages.src	0	string	RID_SVXPAGE_NUM_OPTIONS	STR_BULLET			0	en-US	size *2 \\langle x \\rangle				20050924 09:13:58"
        oofile = self.ooregen(oosource)
        assert r"size *2 \\langle x \\rangle" in oofile



#!/usr/bin/env python
#
# Copyright 2008 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os

from tools.mozilla import moz_l10n_builder


MOZDIR = os.path.join(os.path.expanduser("~"), "mozbuild")


def build_langs(langs, verbose):
    olddir = os.getcwd()
    os.chdir(MOZDIR)

    moz_l10n_builder.main(
        langs=langs,
        mozcheckout=True,
        recover=True,
        potpack=True,
        potincl=["README.mozilla-pot"],
        popack=True,
        update_trans=True,
        diff=False,
        langpack=True,
        verbose=verbose,
    )

    os.chdir(olddir)


def check_potpacks():
    """Copy new and check available POT-packs."""
    pass


def update_rss():
    """Update the RSS feed with the available POT-packs."""
    pass


USAGE = "%prog [<options>]"


def create_option_parser():
    """Creates and returns cmd-line option parser."""

    from argparse import ArgumentParser

    parser = ArgumentParser(usage=USAGE)
    parser.add_argument(
        "-q",
        "--quiet",
        dest="verbose",
        action="store_false",
        default=True,
        help="Print as little as possible output.",
    )

    return parser


def main(langs, verbose):
    if not langs:
        langs = ["ALL"]

    if not os.path.isdir(MOZDIR):
        os.makedirs(MOZDIR)

    build_langs(langs, verbose)
    check_potpacks()
    update_rss()


def main_cmd_line():
    """Processes command-line arguments and send them to main()."""
    args, langs = create_option_parser().parse_args()

    main(langs, args.verbose)


if __name__ == "__main__":
    main_cmd_line()



#
# Copyright 2002-2014 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Convert Java/Mozilla .properties files to Gettext PO localization files.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/prop2po.html
for examples and usage instructions.
"""

import logging

from translate.convert.accesskey import UnitMixer
from translate.storage import po, properties


logger = logging.getLogger(__name__)


class DiscardUnit(ValueError):
    pass


class prop2po:
    """convert a .properties file to a .po file for handling the translation."""

    def __init__(self, personality="java", blankmsgstr=False, duplicatestyle="msgctxt"):
        self.personality = personality
        self.blankmsgstr = blankmsgstr
        self.duplicatestyle = duplicatestyle
        self.mixedkeys = {}
        self.mixer = UnitMixer(properties.labelsuffixes, properties.accesskeysuffixes)

    def convertstore(self, thepropfile):
        """converts a .properties file to a .po file..."""
        thetargetfile = po.pofile()
        if self.personality in ("mozilla", "skype"):
            targetheader = thetargetfile.init_headers(
                x_accelerator_marker="&",
                x_merge_on="location",
            )
        else:
            targetheader = thetargetfile.header()
        targetheader.addnote("extracted from %s" % thepropfile.filename, "developer")

        thepropfile.makeindex()
        self.mixedkeys = self.mixer.match_entities(thepropfile.id_index)
        # we try and merge the header po with any comments at the start of the
        # properties file
        appendedheader = False
        waitingcomments = []
        for propunit in thepropfile.units:
            try:
                pounit = self.convertpropunit(thepropfile, propunit, "developer")
            except DiscardUnit:
                continue
            if pounit is None:
                waitingcomments.extend(propunit.comments)
            if not appendedheader:
                if propunit.isblank():
                    targetheader.addnote(
                        "\n".join(waitingcomments).rstrip(),
                        "developer",
                        position="prepend",
                    )
                    waitingcomments = []
                    pounit = None
                appendedheader = True
            if pounit is not None:
                pounit.addnote(
                    "\n".join(waitingcomments).rstrip(), "developer", position="prepend"
                )
                waitingcomments = []
                thetargetfile.addunit(pounit)
        if self.personality == "gaia":
            thetargetfile = self.fold_gaia_plurals(thetargetfile)
        elif self.personality == "gwt":
            thetargetfile = self.fold_gwt_plurals(thetargetfile)
        thetargetfile.removeduplicates(self.duplicatestyle)
        return thetargetfile

    def mergestore(self, origpropfile, translatedpropfile):
        """converts two .properties files to a .po file..."""
        thetargetfile = po.pofile()
        if self.personality in ("mozilla", "skype"):
            targetheader = thetargetfile.init_headers(
                x_accelerator_marker="&",
                x_merge_on="location",
            )
        else:
            targetheader = thetargetfile.header()
        targetheader.addnote(
            "extracted from %s, %s"
            % (origpropfile.filename, translatedpropfile.filename),
            "developer",
        )
        origpropfile.makeindex()
        # TODO: self.mixedkeys is overwritten below, so this is useless:
        self.mixedkeys = self.mixer.match_entities(origpropfile.id_index)
        translatedpropfile.makeindex()
        self.mixedkeys = self.mixer.match_entities(translatedpropfile.id_index)
        # we try and merge the header po with any comments at the start of
        # the properties file
        appendedheader = False
        waitingcomments = []
        # loop through the original file, looking at units one by one
        for origprop in origpropfile.units:
            try:
                origpo = self.convertpropunit(origpropfile, origprop, "developer")
            except DiscardUnit:
                continue
            if origpo is None:
                waitingcomments.extend(origprop.comments)
            # handle the header case specially...
            if not appendedheader:
                if origprop.isblank():
                    targetheader.addnote(
                        "".join(waitingcomments).rstrip(),
                        "developer",
                        position="prepend",
                    )
                    waitingcomments = []
                    origpo = None
                appendedheader = True
            # try and find a translation of the same name...
            if origprop.name in translatedpropfile.locationindex:
                translatedprop = translatedpropfile.locationindex[origprop.name]
                # Need to check that this comment is not a copy of the
                # developer comments
                try:
                    translatedpo = self.convertpropunit(
                        translatedpropfile, translatedprop, "translator"
                    )
                except DiscardUnit:
                    continue
            else:
                translatedpo = None
            # if we have a valid po unit, get the translation and add it...
            if origpo is not None:
                if translatedpo is not None and not self.blankmsgstr:
                    origpo.target = translatedpo.source
                origpo.addnote(
                    "".join(waitingcomments).rstrip(), "developer", position="prepend"
                )
                waitingcomments = []
                thetargetfile.addunit(origpo)
            elif translatedpo is not None:
                logger.error(
                    "didn't convert original property definition '%s'", origprop.name
                )
        if self.personality == "gaia":
            thetargetfile = self.fold_gaia_plurals(thetargetfile)
        elif self.personality == "gwt":
            thetargetfile = self.fold_gwt_plurals(thetargetfile)
        thetargetfile.removeduplicates(self.duplicatestyle)
        return thetargetfile

    def fold_gwt_plurals(self, postore):
        """Fold the multiple plural units of a gwt file into a gettext plural."""

        def _append_plural_unit(plural_unit, units):
            sources = [u.source for u in units]
            targets = [u.target for u in units]
            # TODO: only consider the right ones for sources and targets
            plural_unit.source = sources
            plural_unit.target = targets
            plural_unit.addlocation(key)

        # Map GWT variants to cldr names
        gwt2cldr = {
            "none": "zero",
            "one": "one",
            "two": "two",
            "few": "few",
            "many": "many",
            "": "other",
        }

        class Variants:
            def __init__(self, unit):
                self.unit = unit
                self.variants = {}

        import re

        from translate.lang import data

        regex = re.compile(r"([^\[\]]*)(?:\[(.*)\])?")
        names = data.cldr_plural_categories
        new_store = type(postore)()
        plurals = {}
        for unit in postore.units:
            if not unit.istranslatable():
                # TODO: reconsider: we could lose header comments here
                continue
            string = unit.getlocations()[0]
            match = regex.match(string)
            if not match:
                logger.warn("Invalid key: %s", string)
                continue
            key = match.group(1)
            variant = match.group(2)
            if key not in plurals:
                # Generate fake unit for each keys
                new_unit = new_store.addsourceunit("fish")
                plurals[key] = Variants(new_unit)

            # No variant => other
            if not variant:
                variant = ""

            # Translate gwt variants to cldr names
            old_variant = variant
            variant = gwt2cldr.get(variant)

            # Some sanity checks
            if not variant:
                raise Exception("Variant invalid: %s" % (old_variant))
            if variant in plurals[key].variants:
                logger.warn(
                    "Override %s[%s]: %s by %s",
                    key,
                    variant,
                    plurals[key].variants[variant],
                    unit,
                )

            # Put the unit
            plurals[key].variants[variant] = unit

        # Rework the set
        for key, plural in plurals.items():
            # We should have at least "other" (no variant in GWT)
            if "other" not in plural.variants:
                raise Exception("Should have property %s without any variant" % (key))
            units = []
            for name in names:
                if name in plural.variants:
                    unit = plural.variants[name]
                    unit.target = unit.source
                    units.append(unit)
            # Replace the sources by good ones
            if "one" in plural.variants and len(units) > 0:
                units[0].source = plural.variants["one"].source
            if "other" in plural.variants and len(units) > 1:
                units[1].source = plural.variants["other"].source

            # Create the plural unit
            _append_plural_unit(plural.unit, units)
        return new_store

    @staticmethod
    def fold_gaia_plurals(postore):
        """Fold the multiple plural units of a gaia file into a gettext plural."""

        def _append_plural_unit(store, plurals, plural):
            units = plurals[plural]
            sources = [u.source for u in units]
            targets = [u.target for u in units]
            # TODO: only consider the right ones for sources and targets
            plural_unit = store.addsourceunit(sources)
            plural_unit.target = targets
            plural_unit.addlocation(plural)
            del plurals[plural]

        new_store = type(postore)()
        plurals = {}
        current_plural = ""
        for unit in postore.units:
            if not unit.istranslatable():
                # TODO: reconsider: we could lose header comments here
                continue
            if "plural(n)" in unit.source:
                if current_plural:
                    # End of a set of plural units
                    _append_plural_unit(new_store, plurals, current_plural)
                    current_plural = ""
                # start of a set of plural units
                location = unit.getlocations()[0]
                current_plural = location
                plurals[location] = []
                # We ignore the first one, since it doesn't contain translatable
                # text, only a marker.
            else:
                location = unit.getlocations()[0]
                if current_plural and location.startswith(current_plural):
                    plurals[current_plural].append(unit)
                    if "[zero]" not in location:
                        # We want to keep [zero] cases separately translatable
                        continue
                elif current_plural:
                    # End of a set of plural units
                    _append_plural_unit(new_store, plurals, current_plural)
                    current_plural = ""

                new_store.addunit(unit)

        if current_plural:
            # The file ended with a set of plural units
            _append_plural_unit(new_store, plurals, current_plural)
            current_plural = ""

        # if everything went well, there should be nothing left in plurals
        if len(plurals) != 0:
            logger.warning(
                "Not all plural units converted correctly:" + "\n".join(plurals)
            )
        return new_store

    @staticmethod
    def convertunit(propunit, commenttype):
        """Converts a .properties unit to a .po unit. Returns None if empty or
        not for translation.
        """
        if propunit is None:
            return None
        # escape unicode
        pounit = po.pounit(encoding="UTF-8")
        if hasattr(propunit, "comments"):
            for comment in propunit.comments:
                if "DONT_TRANSLATE" in comment:
                    raise DiscardUnit(comment)
            pounit.addnote(propunit.getnotes().rstrip(), commenttype)
        # TODO: handle multiline msgid
        if propunit.isblank():
            return None
        pounit.addlocation(propunit.name)
        pounit.source = propunit.source
        pounit.target = ""
        return pounit

    def convertmixedunit(self, labelprop, accesskeyprop, commenttype):
        label_unit = self.convertunit(labelprop, commenttype)
        accesskey_unit = self.convertunit(accesskeyprop, commenttype)
        if label_unit is None:
            return accesskey_unit
        if accesskey_unit is None:
            return label_unit
        target_unit = po.pounit(encoding="UTF-8")
        return self.mixer.mix_units(label_unit, accesskey_unit, target_unit)

    def convertpropunit(self, store, unit, commenttype, mixbucket="properties"):
        """Converts a unit from store to a po unit, keeping track of mixed
        names along the way.

        ``mixbucket`` can be specified to indicate if the given unit is part of
        the template or the translated file.
        """
        if self.personality not in ("mozilla", "gwt"):
            # XXX should we enable unit mixing for other personalities?
            return self.convertunit(unit, commenttype)

        # keep track of whether accesskey and label were combined
        key = unit.getid()
        if key not in self.mixedkeys:
            return self.convertunit(unit, commenttype)

        # use special convertmixed unit which produces one pounit with
        # both combined for the label and None for the accesskey
        alreadymixed = self.mixedkeys[key].get(mixbucket, None)
        if alreadymixed:
            # we are successfully throwing this away...
            return None
        elif alreadymixed is False:
            # The mix failed before
            return self.convertunit(unit, commenttype)

        # assert alreadymixed is None
        labelkey, accesskeykey = self.mixer.find_mixed_pair(self.mixedkeys, store, unit)
        labelprop = store.id_index.get(labelkey, None)
        accesskeyprop = store.id_index.get(accesskeykey, None)
        po_unit = self.convertmixedunit(labelprop, accesskeyprop, commenttype)
        if po_unit is not None:
            if accesskeykey is not None:
                self.mixedkeys[accesskeykey][mixbucket] = True
            if labelkey is not None:
                self.mixedkeys[labelkey][mixbucket] = True
            return po_unit
        else:
            # otherwise the mix failed. add each one separately and
            # remember they weren't mixed
            if accesskeykey is not None:
                self.mixedkeys[accesskeykey][mixbucket] = False
            if labelkey is not None:
                self.mixedkeys[labelkey][mixbucket] = False

        return self.convertunit(unit, commenttype)


def convertstrings(
    inputfile,
    outputfile,
    templatefile,
    personality="strings",
    pot=False,
    duplicatestyle="msgctxt",
    encoding=None,
):
    """.strings specific convertor function"""
    return convertprop(
        inputfile,
        outputfile,
        templatefile,
        personality="strings",
        pot=pot,
        duplicatestyle=duplicatestyle,
        encoding=encoding,
    )


def convertmozillaprop(
    inputfile, outputfile, templatefile, pot=False, duplicatestyle="msgctxt"
):
    """Mozilla specific convertor function"""
    return convertprop(
        inputfile,
        outputfile,
        templatefile,
        personality="mozilla",
        pot=pot,
        duplicatestyle=duplicatestyle,
    )


def convertprop(
    inputfile,
    outputfile,
    templatefile,
    personality="java",
    pot=False,
    duplicatestyle="msgctxt",
    encoding=None,
):
    """reads in inputfile using properties, converts using prop2po, writes to
    outputfile
    """
    inputstore = properties.propfile(inputfile, personality, encoding)
    convertor = prop2po(
        personality=personality, blankmsgstr=pot, duplicatestyle=duplicatestyle
    )
    if templatefile is None:
        outputstore = convertor.convertstore(inputstore)
    else:
        templatestore = properties.propfile(templatefile, personality, encoding)
        outputstore = convertor.mergestore(templatestore, inputstore)
    if outputstore.isempty():
        return 0
    outputstore.serialize(outputfile)
    return 1


formats = {
    "properties": ("po", convertprop),
    ("properties", "properties"): ("po", convertprop),
    "lang": ("po", convertprop),
    ("lang", "lang"): ("po", convertprop),
    "strings": ("po", convertstrings),
    ("strings", "strings"): ("po", convertstrings),
}


def main(argv=None):
    from translate.convert import convert

    parser = convert.ConvertOptionParser(
        formats, usetemplates=True, usepots=True, description=__doc__
    )
    parser.add_option(
        "",
        "--personality",
        dest="personality",
        default=properties.default_dialect,
        type="choice",
        choices=list(properties.dialects.keys()),
        help="override the input file format: %s (for .properties files, default: %s)"
        % (", ".join(properties.dialects.keys()), properties.default_dialect),
        metavar="TYPE",
    )
    parser.add_option(
        "",
        "--encoding",
        dest="encoding",
        default=None,
        help="override the encoding set by the personality",
        metavar="ENCODING",
    )
    parser.add_duplicates_option()
    parser.passthrough.append("pot")
    parser.passthrough.append("personality")
    parser.passthrough.append("encoding")
    parser.run(argv)


if __name__ == "__main__":
    main()



#
# Copyright 2010, 2016 Zuza Software Foundation
#
# This file is part of Virtaal.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""This module represents the Tamil language.

.. seealso:: :wp:`Tamil_language`
"""


from translate.lang import common


class ta(common.Common):
    """This class represents Tamil."""

    ignoretests = {
        "all": ["simplecaps", "startcaps"],
    }



from translate.convert import moz2po, test_convert


class TestMoz2PO:
    pass


class TestMoz2POCommand(test_convert.TestConvertCommand, TestMoz2PO):
    """Tests running actual moz2po commands on files"""

    convertmodule = moz2po
    defaultoptions = {"progress": "none"}

    expected_options = [
        "--duplicates=DUPLICATESTYLE",
        "-P, --pot",
        "-t TEMPLATE, --template=TEMPLATE",
    ]



from translate.misc import quote


def test_find_all():
    """tests the find_all function"""
    assert quote.find_all("", "a") == []
    assert quote.find_all("a", "b") == []
    assert quote.find_all("a", "a") == [0]
    assert quote.find_all("aa", "a") == [0, 1]
    assert quote.find_all("abba", "ba") == [2]
    # check we skip the whole instance
    assert quote.find_all("banana", "ana") == [1]


def test_extract():
    """tests the extract function"""
    assert quote.extract("the <quoted> part", "<", ">", "\\", 0) == ("<quoted>", False)
    assert quote.extract("the 'quoted' part", "'", "'", "\\", 0) == ("'quoted'", False)
    assert quote.extract("the 'isn\\'t escaping fun' part", "'", "'", "\\", 0) == (
        "'isn\\'t escaping fun'",
        False,
    )
    assert quote.extract("the 'isn\\'t something ", "'", "'", "\\", 0) == (
        "'isn\\'t something ",
        True,
    )
    assert quote.extract("<quoted>\\", "<", ">", "\\", 0) == ("<quoted>", False)
    assert quote.extract("<quoted><again>", "<", ">", "\\", 0) == (
        "<quoted><again>",
        False,
    )
    assert quote.extract("<quoted>\\\\<again>", "<", ">", "\\", 0) == (
        "<quoted><again>",
        False,
    )
    assert quote.extract("<quoted\\>", "<", ">", "\\", 0) == ("<quoted\\>", True)
    assert quote.extract(' -->\n<!ENTITY blah "Some">', "<!--", "-->", None, 1) == (
        " -->",
        False,
    )
    assert quote.extract('">\n', '"', '"', None, True) == ('"', False)


def test_extractwithoutquotes():
    """tests the extractwithoutquotes function"""
    assert quote.extractwithoutquotes("the <quoted> part", "<", ">", "\\", 0) == (
        "quoted",
        False,
    )
    assert quote.extractwithoutquotes("the 'quoted' part", "'", "'", "\\", 0) == (
        "quoted",
        False,
    )
    assert quote.extractwithoutquotes(
        "the 'isn\\'t escaping fun' part", "'", "'", "\\", 0
    ) == ("isn\\'t escaping fun", False)
    assert quote.extractwithoutquotes("the 'isn\\'t something ", "'", "'", "\\", 0) == (
        "isn\\'t something ",
        True,
    )
    assert quote.extractwithoutquotes("<quoted>\\", "<", ">", "\\", 0) == (
        "quoted",
        False,
    )
    assert quote.extractwithoutquotes("<quoted>\\\\<again>", "<", ">", "\\", 0) == (
        "quotedagain",
        False,
    )
    assert quote.extractwithoutquotes(
        "<quoted><again\\\\", "<", ">", "\\", 0, True
    ) == ("quotedagain\\\\", True)
    # don't include escapes...
    assert quote.extractwithoutquotes(
        "the 'isn\\'t escaping fun' part", "'", "'", "\\", 0, False
    ) == ("isn't escaping fun", False)
    assert quote.extractwithoutquotes(
        "the 'isn\\'t something ", "'", "'", "\\", 0, False
    ) == ("isn't something ", True)
    assert quote.extractwithoutquotes("<quoted\\", "<", ">", "\\", 0, False) == (
        "quoted",
        True,
    )
    assert quote.extractwithoutquotes(
        "<quoted><again\\\\", "<", ">", "\\", 0, False
    ) == ("quotedagain\\", True)
    # escaping of quote char
    assert quote.extractwithoutquotes("<quoted\\>", "<", ">", "\\", 0, False) == (
        "quoted>",
        True,
    )


def isnewlineortabescape(escape):
    if escape in ("\\n", "\\t"):
        return escape
    return escape[-1]


def test_extractwithoutquotes_passfunc():
    """tests the extractwithoutquotes function with a function for includeescapes as a parameter"""
    assert quote.extractwithoutquotes(
        "<test \\r \\n \\t \\\\>", "<", ">", "\\", 0, isnewlineortabescape
    ) == ("test r \\n \\t \\", False)


def test_stripcomment():
    assert quote.stripcomment("<!-- Comment -->") == "Comment"


class TestEncoding:
    @staticmethod
    def test_javapropertiesencode():
        assert quote.javapropertiesencode("abc") == "abc"
        assert quote.javapropertiesencode("abcḓ") == r"abc\u1E13"
        assert quote.javapropertiesencode("abc\n") == "abc\\n"

    @staticmethod
    def test_java_utf8_properties_encode():
        assert quote.java_utf8_properties_encode("abc") == "abc"
        assert quote.java_utf8_properties_encode("abcḓ") == "abcḓ"
        assert quote.java_utf8_properties_encode("abc\n") == "abc\\n"

    @staticmethod
    def test_escapespace():
        assert quote.escapespace(" ") == "\\u0020"
        assert quote.escapespace("\t") == "\\u0009"

    @staticmethod
    def test_mozillaescapemarginspaces():
        assert quote.mozillaescapemarginspaces(" ") == ""
        assert quote.mozillaescapemarginspaces("A") == "A"
        assert quote.mozillaescapemarginspaces(" abc ") == "\\u0020abc\\u0020"
        assert quote.mozillaescapemarginspaces("  abc ") == "\\u0020 abc\\u0020"

    @staticmethod
    def test_mozilla_control_escapes():
        r"""test that we do \uNNNN escapes for certain control characters instead of converting to UTF-8 characters"""
        prefix, suffix = "bling", "blang"
        for control in ("\u0005", "\u0006", "\u0007", "\u0011"):
            string = prefix + control + suffix
            assert quote.escapecontrols(string) == string

    @staticmethod
    def test_propertiesdecode():
        assert quote.propertiesdecode("abc") == "abc"
        assert quote.propertiesdecode("abc\\u1e13") == "abcḓ"
        assert quote.propertiesdecode("abc\\u1E13") == "abcḓ"
        assert quote.propertiesdecode("abc\N{LEFT CURLY BRACKET}") == "abc{"
        assert quote.propertiesdecode("abc\\") == "abc\\"
        assert quote.propertiesdecode("abc\\") == "abc\\"

    @staticmethod
    def test_properties_decode_slashu():
        # The real input strings don't have double backslashes, but we have to
        # double them here because Python immediately decode them, even for raw
        # strings.
        assert quote.propertiesdecode("abc\\u1e13") == "abcḓ"
        assert quote.propertiesdecode("abc\\u0020") == "abc "
        # NOTE Java only accepts 4 digit unicode, Mozilla accepts two
        # unfortunately, but it seems harmless to accept both.
        assert quote.propertiesdecode("abc\\u20") == "abc "

    @staticmethod
    def _html_encoding_helper(pairs):
        for from_, to in pairs:
            assert quote.htmlentityencode(from_) == to
            assert quote.htmlentitydecode(to) == from_

    def test_htmlencoding(self):
        """test that we can encode and decode simple HTML entities"""
        raw_encoded = [("€", "&euro;"), ("©", "&copy;"), ('"', "&quot;")]
        self._html_encoding_helper(raw_encoded)

    @staticmethod
    def test_htmlencoding_existing_entities():
        """test that we don't mess existing entities"""
        assert quote.htmlentityencode("&amp;") == "&amp;"

    def test_htmlencoding_passthrough(self):
        """test that we can encode and decode things that look like HTML entities but aren't"""
        raw_encoded = [
            ("copy quot", "copy quot")
        ]  # Raw text should have nothing done to it.
        self._html_encoding_helper(raw_encoded)

    @staticmethod
    def test_htmlencoding_nonentities():
        """tests to give us full coverage"""
        for encoded, real in [
            ("Some &; text", "Some &; text"),
            ("&copy ", "&copy "),
            ("&copy", "&copy"),
            ("&rogerrabbit;", "&rogerrabbit;"),
        ]:
            assert quote.htmlentitydecode(encoded) == real

        for decoded, real in [
            ("Some &; text", "Some &; text"),
            ("&copy ", "&amp;copy "),
            ("&copy", "&amp;copy"),
            ("&rogerrabbit;", "&rogerrabbit;"),
        ]:
            assert quote.htmlentityencode(decoded) == real



from io import BytesIO

import pytest

from translate.convert import test_convert, txt2po


class BaseTxt2POTester:

    ConverterClass = txt2po.txt2po
    Flavour = None

    def _convert(
        self,
        input_string,
        template_string=None,
        duplicate_style="msgctxt",
        encoding="utf-8",
        success_expected=True,
        no_segmentation=False,
    ):
        """Helper that converts to target format without using files."""
        input_file = BytesIO(input_string.encode())
        output_file = BytesIO()
        template_file = None
        if template_string:
            template_file = BytesIO(template_string.encode())
        expected_result = 1 if success_expected else 0
        converter = self.ConverterClass(
            input_file,
            output_file,
            template_file,
            duplicate_style,
            encoding,
            self.Flavour,
            no_segmentation,
        )
        assert converter.run() == expected_result
        return converter.target_store, output_file

    def _convert_to_store(self, *args, **kwargs):
        """Helper that converts to target format store without using files."""
        return self._convert(*args, **kwargs)[0]

    def _convert_to_string(self, *args, **kwargs):
        """Helper that converts to target format string without using files."""
        return self._convert(*args, **kwargs)[1].getvalue().decode("utf-8")

    @staticmethod
    def _count_elements(po_store):
        """Helper that counts the number of non-header units."""
        assert po_store.units[0].isheader()
        return len(po_store.units) - 1


class TestTxt2PO(BaseTxt2POTester):
    def test_convert_empty(self):
        """Check converting empty file returns no output."""
        assert self._convert_to_string("", success_expected=False) == ""

    def test_keep_duplicates(self):
        """Check converting keeps duplicates."""
        input_string = """
Simple

Simple
"""
        target_store = self._convert_to_store(input_string)
        assert self._count_elements(target_store) == 2
        assert target_store.units[1].source == "Simple"
        assert target_store.units[1].target == ""
        assert target_store.units[2].source == "Simple"
        assert target_store.units[2].target == ""

    def test_drop_duplicates(self):
        """Check converting drops duplicates."""
        input_string = """
Simple

Simple
"""
        target_store = self._convert_to_store(input_string, duplicate_style="merge")
        assert self._count_elements(target_store) == 1
        assert target_store.units[1].source == "Simple"
        assert target_store.units[1].target == ""

    def test_simple(self):
        """Test the most basic conversion."""
        input_string = "A simple string"
        expected_output = """#: :1
msgid "A simple string"
msgstr ""
"""
        target_store = self._convert_to_store(input_string)
        assert str(target_store.units[1]) == expected_output
        assert "extracted from " in str(target_store.header())

    def test_multiple_units(self):
        """Test that we can handle txt with multiple units."""
        input_string = """First unit
Still part of first unit

Second unit is a heading
------------------------

Third unit with blank after but no more units.

"""
        target_store = self._convert_to_store(input_string)
        assert self._count_elements(target_store) == 3

    def test_carriage_return(self):
        """Remove carriage returns from files in dos format."""
        input_string = """The rapid expansion of telecommunications infrastructure in recent years has\r
helped to bridge the digital divide to a limited extent.\r
"""
        expected_output = """The rapid expansion of telecommunications infrastructure in recent years has
helped to bridge the digital divide to a limited extent."""

        target_store = self._convert_to_store(input_string)
        assert str(target_store.units[1].source) == expected_output

    def test_merge(self):
        """Test converter doesn't merge."""
        with pytest.raises(NotImplementedError):
            self._convert_to_store("this", "cannot be", "blank", success_expected=False)

    def test_no_segmentation(self):
        """Check multiple paragraphs are extracted as a single unit."""
        input_string = """
First paragraph

Second paragraph
"""
        expected_output = input_string
        target_store = self._convert_to_store(input_string, no_segmentation=True)
        assert self._count_elements(target_store) == 1
        assert target_store.units[1].source == expected_output
        assert target_store.units[1].target == ""


class TestDoku2po(BaseTxt2POTester):

    Flavour = "dokuwiki"

    def test_convert_empty(self):
        """Test converting empty file returns no output."""
        assert self._convert_to_string("", success_expected=False) == ""

    def test_keep_duplicates(self):
        """Check converting keeps duplicates."""
        input_string = """
Simple

Simple
"""
        target_store = self._convert_to_store(input_string)
        assert self._count_elements(target_store) == 2
        assert target_store.units[1].source == "Simple"
        assert target_store.units[1].target == ""
        assert target_store.units[2].source == "Simple"
        assert target_store.units[2].target == ""

    def test_drop_duplicates(self):
        """Test converting drops duplicates."""
        input_string = """
Simple

Simple
"""
        target_store = self._convert_to_store(input_string, duplicate_style="merge")
        assert self._count_elements(target_store) == 1
        assert target_store.units[1].source == "Simple"
        assert target_store.units[1].target == ""

    def test_basic(self):
        """Test basic Dokuwiki conversion."""
        input_string = """=====Heading=====

This is a wiki page.
"""
        target_store = self._convert_to_store(input_string)
        assert self._count_elements(target_store) == 2
        assert target_store.units[1].source == "Heading"
        assert target_store.units[2].source == "This is a wiki page."

    def test_bullet_list(self):
        """Test Dokuwiki bullet list conversion."""
        input_string = """  * This is a fact.
  * This is a fact.
"""
        target_store = self._convert_to_store(input_string)
        assert self._count_elements(target_store) == 2
        assert target_store.units[1].source == "This is a fact."
        assert target_store.units[1].getlocations() == [":1"]
        assert target_store.units[2].source == "This is a fact."
        assert target_store.units[2].getlocations() == [":2"]

    def test_numbered_list(self):
        """Test Dokuwiki numbered list conversion."""
        input_string = """  - This is an item.
  - This is an item.
"""
        target_store = self._convert_to_store(input_string)
        assert self._count_elements(target_store) == 2
        assert target_store.units[1].source == "This is an item."
        assert target_store.units[1].getlocations() == [":1"]
        assert target_store.units[2].source == "This is an item."
        assert target_store.units[2].getlocations() == [":2"]

    def test_spacing(self):
        """Test Dokuwiki list nesting conversion."""
        input_string = """ =====         Heading  =====
  * This is an item.
    * This is a subitem.
        * This is a tabbed item.
"""
        target_store = self._convert_to_store(input_string)
        assert self._count_elements(target_store) == 4
        assert target_store.units[1].source == "Heading"
        assert target_store.units[2].source == "This is an item."
        assert target_store.units[3].source == "This is a subitem."
        assert target_store.units[4].source == "This is a tabbed item."

    def test_merge(self):
        """Test converter doesn't merge."""
        with pytest.raises(NotImplementedError):
            self._convert_to_store("this", "cannot be", "blank", success_expected=False)


class TestTxt2POCommand(test_convert.TestConvertCommand, TestTxt2PO):
    """Tests running actual txt2po commands on files"""

    convertmodule = txt2po
    defaultoptions = {"progress": "none"}
    expected_options = [
        "-P, --pot",
        "--duplicates",
        "--encoding",
        "--flavour",
        "--no-segmentation",
    ]



from translate.convert import rc2po, test_convert
from translate.storage.po import pofile


RC_SOURCE = r"""
#include "other_file.h" // This must be ignored

LANGUAGE LANG_ENGLISH, SUBLANG_DEFAULT

/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_REGGHC_DIALOG DIALOGEX 0, 0, 211, 191
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
EXSTYLE WS_EX_APPWINDOW
CAPTION "License dialog"
FONT 8, "MS Shell Dlg", 0, 0, 0x1
BEGIN
    PUSHBUTTON      "Help",ID_HELP,99,162,48,15
    PUSHBUTTON      "Close",IDCANCEL,151,162,48,15
    PUSHBUTTON      "Activate instalation",IDC_BUTTON1,74,76,76,18
    CTEXT           "My very good program",IDC_STATIC1,56,21,109,19,SS_SUNKEN
    CTEXT           "You can use it without registering it",IDC_STATIC,35,131,128,19,SS_SUNKEN
    PUSHBUTTON      "Offline",IDC_OFFLINE,149,108,42,13
    PUSHBUTTON      "See license",IDC_LICENCIA,10,162,85,15
    RTEXT           "If you don't have internet, please use magic.",IDC_STATIC,23,105,120,18
    ICON            IDR_MAINFRAME,IDC_STATIC,44,74,20,20
    CTEXT           "Use your finger to activate the program.",IDC_ACTIVADA,17,50,175,17
    ICON            IDR_MAINFRAME1,IDC_STATIC6,18,19,20,20
END

MainMenu MENU
{
    POPUP "&Debug"
    {
        MENUITEM "&Memory usage", ID_MEMORY
        POPUP
        {
            MENUITEM SEPARATOR
            MENUITEM "&Walk data heap", ID_WALK_HEAP
        }
    }
}

STRINGTABLE
BEGIN
ID_T_1 "Hello"
END

IDR_MAINFRAME MENU
BEGIN
   POPUP "&File"
   BEGIN
       MENUITEM "&New\tCrtl+N",                ID_FILE_NEW
       MENUITEM "&Open...\tCtrl+O",            ID_FILE_OPEN
       MENUITEM "&Exit",                       ID_FILE_CLOSE
   END
   POPUP "&View"
   BEGIN
       MENUITEM "&Toolbar",                    ID_VIEW_STATUS_BAR
   END
   POPUP "&Help"
   BEGIN
       MENUITEM "&About...",                   ID_APP_ABOUT
   END
END
"""


class TestRC2POCommand(test_convert.TestConvertCommand):
    """Tests running actual rc2po commands on files"""

    convertmodule = rc2po
    defaultoptions = {"progress": "none"}
    expected_options = [
        "-t TEMPLATE, --template=TEMPLATE",
        "-l LANG, --lang=LANG",
        "-P, --pot",
        "--charset=CHARSET",
        "--sublang=SUBLANG",
        "--duplicates=DUPLICATESTYLE",
    ]

    def test_convert(self):
        """Tests the conversion to a po file"""
        self.create_testfile("simple.rc", RC_SOURCE)
        self.run_command(i="simple.rc", o="simple.po")
        po_result = pofile(self.open_testfile("simple.po"))
        assert len(po_result.units) == 23
        # first unit is PO file header
        assert po_result.units[1].source == "License dialog"
        assert po_result.units[11].source == "&Debug"
        assert po_result.units[12].source == "&Memory usage"
        assert po_result.units[13].source == "&Walk data heap"
        assert po_result.units[14].source == "Hello"
        assert po_result.units[15].source == "&File"
        assert po_result.units[15].getlocations() == [
            "MENU.IDR_MAINFRAME.POPUP.CAPTION"
        ]

    def test_convert_encoding_utf16(self):
        self.create_testfile("simple.rc", RC_SOURCE.encode("utf-16"))
        self.run_command(i="simple.rc", o="simple.po", charset="utf-16")
        po_result = pofile(self.open_testfile("simple.po"))
        assert len(po_result.units) == 23

    def test_convert_encoding_wrong(self):
        self.create_testfile("simple.rc", RC_SOURCE.encode("utf-8"))
        self.run_command(i="simple.rc", o="simple.po", charset="utf-16x")
        po_result = pofile(self.open_testfile("simple.po"))
        assert len(po_result.units) == 0
        self.run_command(i="simple.rc", o="simple.po", charset="utf-16")
        po_result = pofile(self.open_testfile("simple.po"))
        assert len(po_result.units) == 0

    def test_convert_encoding_utf8(self):
        self.create_testfile("simple.rc", RC_SOURCE.encode("utf-8"))
        self.run_command(i="simple.rc", o="simple.po", charset="utf-8")
        po_result = pofile(self.open_testfile("simple.po"))
        assert len(po_result.units) == 23

    def test_menuex(self):
        source = """
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US

IDM_STARTMENU MENUEX
BEGIN
    POPUP ""
    BEGIN
        MENUITEM "", -1, MFT_SEPARATOR
        POPUP "&Programs", IDM_PROGRAMS
        BEGIN
            MENUITEM "(Empty)", -1, MFT_STRING, MFS_GRAYED
        END
        MENUITEM "Sh&ut Down...", IDM_SHUTDOWN, MFT_STRING, MFS_ENABLED
    END
END
"""
        self.create_testfile("simple.rc", source.encode("utf-8"))
        self.run_command(i="simple.rc", o="simple.po", charset="utf-8")
        po_result = pofile(self.open_testfile("simple.po"))
        assert len(po_result.units) == 5



from translate.storage import ts


class TestTS:
    @staticmethod
    def test_construct():
        tsfile = ts.QtTsParser()
        tsfile.addtranslation("ryan", "Bread", "Brood", "Wit", createifmissing=True)



#
# Copyright 2007, 2016 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""This module represents the Khmer language.

.. seealso:: :wp:`Khmer_language`
"""


import re

from translate.lang import common


class km(common.Common):
    """This class represents Khmer."""

    khmerpunc = "។៕៖៘"
    """These marks are only used for Khmer."""

    punctuation = "".join(
        [
            common.Common.commonpunc,
            common.Common.quotes,
            common.Common.miscpunc,
            khmerpunc,
        ]
    )

    sentenceend = "!?…។៕៘"

    sentencere = re.compile(
        r"""(?s)    #make . also match newlines
                            .*?         #anything, but match non-greedy
                            [%s]        #the puntuation for sentence ending
                            \s+         #the spacing after the puntuation
                            (?=[^a-z\d])#lookahead that next part starts with caps
                            """
        % sentenceend,
        re.VERBOSE,
    )
    # \u00a0 is non-breaking space
    puncdict = {
        ".": "\u00a0។",
        ":": "\u00a0៖",
        "!": "\u00a0!",
        "?": "\u00a0?",
    }

    ignoretests = {
        "all": ["simplecaps", "startcaps"],
    }

    mozilla_nplurals = 2
    mozilla_pluralequation = "n!=1 ? 1 : 0"



#
# Copyright 2007 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""This module provides functionality to work with directories."""

# Perhaps all methods should work with a wildcard to limit searches in some
# way (examples: *.po, base.xlf, pootle-terminology.tbx)

# TODO: consider also providing directories as we currently provide files

import os

from translate.storage import factory


class Directory:
    """This class represents a directory."""

    def __init__(self, dir=None):
        self.dir = dir
        self.filedata = []

    def file_iter(self):
        """Iterator over (dir, filename) for all files in this directory."""
        if not self.filedata:
            self.scanfiles()
        yield from self.filedata

    def getfiles(self):
        """Returns a list of (dir, filename) tuples for all the file names in
        this directory.
        """
        return list(self.file_iter())

    def unit_iter(self):
        """Iterator over all the units in all the files in this directory."""
        for dirname, filename in self.file_iter():
            store = factory.getobject(os.path.join(dirname, filename))
            # TODO: don't regenerate all the storage objects
            yield from store.unit_iter()

    def getunits(self):
        """List of all the units in all the files in this directory."""
        return list(self.unit_iter())

    def scanfiles(self):
        """Populate the internal file data."""
        self.filedata = []

        for dirpath, dirnames, filenames in os.walk(self.dir):
            fnames = dirnames + filenames
            for fname in fnames:
                if os.path.isfile(os.path.join(dirpath, fname)):
                    self.filedata.append((dirpath, fname))



from translate.lang import factory


def test_punctranslate():
    """Tests that we can translate punctuation."""
    language = factory.getlanguage("km")
    assert language.punctranslate("") == ""
    assert language.punctranslate("abc efg") == "abc efg"
    assert language.punctranslate("abc efg.") == "abc efg\u00a0។"
    print(language.punctranslate("abc efg. hij.").encode("utf-8"))
    print("abc efg\u00a0។ hij\u00a0។".encode())
    assert language.punctranslate("abc efg. hij.") == "abc efg\u00a0។ hij\u00a0។"
    assert language.punctranslate("abc efg!") == "abc efg\u00a0!"
    assert language.punctranslate("abc efg? hij!") == "abc efg\u00a0? hij\u00a0!"
    assert language.punctranslate("Delete file: %s?") == "Delete file\u00a0៖ %s\u00a0?"


def test_sentences():
    """Tests basic functionality of sentence segmentation."""
    language = factory.getlanguage("km")
    sentences = language.sentences("")
    assert sentences == []

    sentences = language.sentences(
        "លក្ខណៈ​​នេះ​អាច​ឲ្យ​យើងធ្វើ​ជាតូបនីយកម្មកម្មវិធី​កុំព្យូទ័រ​ ។ លក្ខណៈ​​នេះ​អាច​ឲ្យ​យើងធ្វើ​ជាតូបនីយកម្មកម្មវិធី​កុំព្យូទ័រ​ ។"
    )
    print(sentences)
    assert sentences == [
        "លក្ខណៈ​​នេះ​អាច​ឲ្យ​យើងធ្វើ​ជាតូបនីយកម្មកម្មវិធី​កុំព្យូទ័រ​ ។",
        "លក្ខណៈ​​នេះ​អាច​ឲ្យ​យើងធ្វើ​ជាតូបនីយកម្មកម្មវិធី​កុំព្យូទ័រ​ ។",
    ]



#
# Copyright 2007-2009 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Convert Windows RC files to Gettext PO localization files.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/rc2po.html
for examples and usage instructions.
"""

import logging

from translate.storage import po, rc


logger = logging.getLogger(__name__)


class rc2po:
    """Convert a .rc file to a .po file for handling the translation."""

    def convert_store(self, input_store, duplicatestyle="msgctxt"):
        """converts a .rc file to a .po file..."""
        output_store = po.pofile()
        output_header = output_store.init_headers(
            x_accelerator_marker="&",
            x_merge_on="location",
        )
        output_header.addnote("extracted from %s" % input_store.filename, "developer")
        for input_unit in input_store.units:
            output_unit = self.convert_unit(input_unit, "developer")
            if output_unit is not None:
                output_store.addunit(output_unit)
        output_store.removeduplicates(duplicatestyle)
        return output_store

    def merge_store(
        self, template_store, input_store, blankmsgstr=False, duplicatestyle="msgctxt"
    ):
        """converts two .rc files to a .po file..."""
        output_store = po.pofile()
        output_header = output_store.init_headers(
            x_accelerator_marker="&",
            x_merge_on="location",
        )
        output_header.addnote(
            f"extracted from {template_store.filename}, {input_store.filename}",
            "developer",
        )
        input_store.makeindex()
        for template_unit in template_store.units:
            origpo = self.convert_unit(template_unit, "developer")
            # try and find a translation of the same name...
            template_unit_name = "".join(template_unit.getlocations())
            if template_unit_name in input_store.locationindex:
                translatedrc = input_store.locationindex[template_unit_name]
                translatedpo = self.convert_unit(translatedrc, "translator")
            else:
                translatedpo = None
            # if we have a valid po unit, get the translation and add it...
            if origpo is not None:
                if translatedpo is not None and not blankmsgstr:
                    origpo.target = translatedpo.source
                output_store.addunit(origpo)
            elif translatedpo is not None:
                logging.error(
                    "error converting original rc definition %s", template_unit.name
                )
        output_store.removeduplicates(duplicatestyle)
        return output_store

    @staticmethod
    def convert_unit(input_unit, commenttype):
        """Converts a .rc unit to a .po unit. Returns None if empty or not for
        translation.
        """
        if input_unit is None:
            return None
        # escape unicode
        output_unit = po.pounit(encoding="UTF-8")
        output_unit.addlocation("".join(input_unit.getlocations()))
        output_unit.source = input_unit.source
        output_unit.target = ""
        return output_unit


def convertrc(
    input_file,
    output_file,
    template_file,
    pot=False,
    duplicatestyle="msgctxt",
    charset=None,
    lang=None,
    sublang=None,
):
    """reads in input_file using rc, converts using rc2po, writes to output_file"""
    input_store = rc.rcfile(input_file, lang, sublang, encoding=charset)
    convertor = rc2po()
    if template_file is None:
        output_store = convertor.convert_store(
            input_store, duplicatestyle=duplicatestyle
        )
    else:
        template_store = rc.rcfile(template_file, lang, sublang, encoding=charset)
        output_store = convertor.merge_store(
            template_store, input_store, blankmsgstr=pot, duplicatestyle=duplicatestyle
        )
    if output_store.isempty():
        return 0
    output_store.serialize(output_file)
    return 1


def main(argv=None):
    from translate.convert import convert

    formats = {
        "rc": ("po", convertrc),
        ("rc", "rc"): ("po", convertrc),
        "nls": ("po", convertrc),
        ("nls", "nls"): ("po", convertrc),
    }
    parser = convert.ConvertOptionParser(
        formats, usetemplates=True, usepots=True, description=__doc__
    )
    parser.add_option(
        "",
        "--charset",
        dest="charset",
        default="auto",
        help="charset to use to decode the RC files (autodetection is used by default)",
        metavar="CHARSET",
    )
    DEFAULTLANG = None
    parser.add_option(
        "-l",
        "--lang",
        dest="lang",
        default=DEFAULTLANG,
        help="LANG entry (default: %s)" % DEFAULTLANG,
        metavar="LANG",
    )
    DEFAULTSUBLANG = None
    parser.add_option(
        "",
        "--sublang",
        dest="sublang",
        default=DEFAULTSUBLANG,
        help="SUBLANG entry (default: %s)" % DEFAULTSUBLANG,
        metavar="SUBLANG",
    )
    parser.add_duplicates_option()
    parser.passthrough.append("pot")
    parser.passthrough.append("charset")
    parser.passthrough.append("lang")
    parser.passthrough.append("sublang")
    parser.run(argv)


if __name__ == "__main__":
    main()



from translate.lang import factory


def test_punctranslate():
    """Tests that we can translate punctuation."""
    language = factory.getlanguage("ar")
    assert language.punctranslate("") == ""
    assert language.punctranslate("abc efg") == "abc efg"
    assert language.punctranslate("abc efg.") == "abc efg."
    assert language.punctranslate("abc, efg; d?") == "abc، efg؛ d؟"
    # See https://github.com/translate/translate/issues/1819
    assert language.punctranslate("It is called “abc”") == "It is called ”abc“"


def test_sentences():
    """Tests basic functionality of sentence segmentation."""
    language = factory.getlanguage("ar")
    sentences = language.sentences("")
    assert sentences == []

    sentences = language.sentences('يوجد بالفعل مجلد بالإسم "%s". أترغب في استبداله؟')
    print(sentences)
    assert sentences == ['يوجد بالفعل مجلد بالإسم "%s".', "أترغب في استبداله؟"]
    # This probably doesn't make sense: it is just the above reversed, to make sure
    # we test the '؟' as an end of sentence marker.
    sentences = language.sentences('أترغب في استبداله؟ يوجد بالفعل مجلد بالإسم "%s".')
    print(sentences)
    assert sentences == ["أترغب في استبداله؟", 'يوجد بالفعل مجلد بالإسم "%s".']



#
# Copyright 2005, 2006 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Convert Gettext PO localization files to a TMX (Translation Memory eXchange) file.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/po2tmx.html
for examples and usage instructions.
"""

import os

from translate.convert import convert
from translate.misc import wStringIO
from translate.storage import po, tmx


class po2tmx:
    @staticmethod
    def cleancomments(comments, comment_type=None):
        """Removes the comment marks from the PO strings."""
        # FIXME this is a bit hacky, needs some fixes in the PO classes
        for index, comment in enumerate(comments):
            if comment.startswith("#"):
                if comment_type is None:
                    comments[index] = comment[1:].rstrip()
                else:
                    comments[index] = comment[2:].strip()

        return "".join(comments)

    def convertfiles(
        self, inputfile, tmxfile, sourcelanguage="en", targetlanguage=None, comment=None
    ):
        """converts a .po file (possibly many) to TMX file"""
        inputstore = po.pofile(inputfile)
        for inunit in inputstore.units:
            if (
                inunit.isheader()
                or inunit.isblank()
                or not inunit.istranslated()
                or inunit.isfuzzy()
            ):
                continue
            source = inunit.source
            translation = inunit.target

            commenttext = {
                "source": self.cleancomments(inunit.sourcecomments, "source"),
                "type": self.cleancomments(inunit.typecomments, "type"),
                "others": self.cleancomments(inunit.othercomments),
            }.get(comment)

            tmxfile.addtranslation(
                source, sourcelanguage, translation, targetlanguage, commenttext
            )


def convertpo(
    inputfile,
    outputfile,
    templatefile,
    sourcelanguage="en",
    targetlanguage=None,
    comment=None,
):
    """reads in stdin using fromfileclass, converts using convertorclass, writes to stdout"""
    convertor = po2tmx()
    convertor.convertfiles(
        inputfile, outputfile.tmxfile, sourcelanguage, targetlanguage, comment
    )
    return 1


class tmxmultifile:
    def __init__(self, filename, mode=None):
        """initialises tmxmultifile from a seekable inputfile or writable outputfile"""
        self.filename = filename
        if mode is None:
            if os.path.exists(filename):
                mode = "r"
            else:
                mode = "w"
        self.mode = mode
        #        self.multifilestyle = multifilestyle
        self.multifilename = os.path.splitext(filename)[0]
        #        self.multifile = open(filename, mode)
        self.tmxfile = tmx.tmxfile()

    def openoutputfile(self, subfile):
        """returns a pseudo-file object for the given subfile"""

        def onclose(contents):
            pass

        outputfile = wStringIO.CatchStringOutput(onclose)
        outputfile.filename = subfile
        outputfile.tmxfile = self.tmxfile
        return outputfile


class TmxOptionParser(convert.ArchiveConvertOptionParser):
    def recursiveprocess(self, options):
        if not options.targetlanguage:
            raise ValueError("You must specify the target language")
        super().recursiveprocess(options)
        with open(options.output, "wb") as self.output:
            self.outputarchive.tmxfile.setsourcelanguage(options.sourcelanguage)
            self.outputarchive.tmxfile.serialize(self.output)


def main(argv=None):
    formats = {"po": ("tmx", convertpo), ("po", "tmx"): ("tmx", convertpo)}
    archiveformats = {(None, "output"): tmxmultifile, (None, "template"): tmxmultifile}
    parser = TmxOptionParser(
        formats,
        usepots=False,
        usetemplates=False,
        description=__doc__,
        archiveformats=archiveformats,
    )
    parser.add_option(
        "-l",
        "--language",
        dest="targetlanguage",
        default=None,
        help="set target language code (e.g. af-ZA) [required]",
        metavar="LANG",
    )
    parser.add_option(
        "",
        "--source-language",
        dest="sourcelanguage",
        default="en",
        help="set source language code (default: en)",
        metavar="LANG",
    )
    comments = ["source", "type", "others", "none"]
    comments_help = (
        "set default comment import: none, source, type or " "others (default: none)"
    )
    parser.add_option(
        "",
        "--comments",
        dest="comment",
        default="none",
        type="choice",
        choices=comments,
        help=comments_help,
    )
    parser.passthrough.append("sourcelanguage")
    parser.passthrough.append("targetlanguage")
    parser.passthrough.append("comment")
    parser.run(argv)


if __name__ == "__main__":
    main()



from translate.filters import autocorrect


class TestAutocorrect:
    @staticmethod
    def correct(msgid, msgstr, expected):
        """helper to run correct function from autocorrect module"""
        corrected = autocorrect.correct(msgid, msgstr)
        print(repr(msgid))
        print(repr(msgstr))
        print(msgid.encode("utf-8"))
        print(msgstr.encode("utf-8"))
        print((corrected or "").encode("utf-8"))
        assert corrected == expected

    def test_empty_target(self):
        """test that we do nothing with an empty target"""
        self.correct("String...", "", None)

    def test_correct_ellipsis(self):
        """test that we convert single … or ... to match source and target"""
        self.correct("String...", "Translated…", "Translated...")
        self.correct("String…", "Translated...", "Translated…")

    def test_correct_spacestart_spaceend(self):
        """test that we can correct leading and trailing space errors"""
        self.correct("Simple string", "Dimpled ring  ", "Dimpled ring")
        self.correct("Simple string", "  Dimpled ring", "Dimpled ring")
        self.correct("  Simple string", "Dimpled ring", "  Dimpled ring")
        self.correct("Simple string  ", "Dimpled ring", "Dimpled ring  ")

    def test_correct_start_capitals(self):
        """test that we can correct the starting capital"""
        self.correct("Simple string", "dimpled ring", "Dimpled ring")
        self.correct("simple string", "Dimpled ring", "dimpled ring")

    def test_correct_end_punc(self):
        """test that we can correct end punctuation"""
        self.correct("Simple string:", "Dimpled ring", "Dimpled ring:")
        # self.correct("Simple string: ", "Dimpled ring", "Dimpled ring: ")
        self.correct("Simple string.", "Dimpled ring", "Dimpled ring.")
        # self.correct("Simple string. ", "Dimpled ring", "Dimpled ring. ")
        self.correct("Simple string?", "Dimpled ring", "Dimpled ring?")

    def test_correct_combinations(self):
        """test that we can correct combinations of failures"""
        self.correct("Simple string:", "Dimpled ring ", "Dimpled ring:")
        self.correct("simple string ", "Dimpled ring", "dimpled ring ")
        self.correct("Simple string...", "dimpled ring..", "Dimpled ring...")
        self.correct("Simple string:", "Dimpled ring ", "Dimpled ring:")

    def test_nothing_to_do(self):
        """test that when nothing changes we return None"""
        self.correct("Some text", "A translation", None)



import os
from io import BytesIO
from urllib import parse

from translate.convert import oo2po, po2oo, test_convert
from translate.storage import oo, po
from translate.storage.poheader import poheader


class TestOO2PO:
    target_filetype = po.pofile
    conversion_module = oo2po
    conversion_class = oo2po.oo2po

    def convert(self, oosource, sourcelanguage="en-US", targetlanguage="af-ZA"):
        """helper that converts oo source to po source without requiring files"""
        if isinstance(oosource, str):
            oosource = oosource.encode("utf-8")
        inputoo = oo.oofile(oosource)
        convertor = self.conversion_class(sourcelanguage, targetlanguage)
        outputpo = convertor.convertstore(inputoo)
        return outputpo

    @staticmethod
    def singleelement(pofile):
        """checks that the pofile contains a single non-header element, and returns it"""
        if isinstance(pofile, poheader):
            assert len(pofile.units) == 2
            assert pofile.units[0].isheader()
            return pofile.units[1]
        else:
            assert len(pofile.units) == 1
            return pofile.units[0]

    def roundtripstring(self, filename, entitystring):
        """
        Convert the supplied string as part of an OpenOffice.org GSI file to po and back.

        Return the string once it has been through all the conversions.
        """

        ootemplate = r"helpcontent2	%s	0	help	par_id3150670 35				0	en-US	%s				2002-02-02 02:02:02"

        oosource = (ootemplate % (filename, entitystring)).encode("utf-8")
        ooinputfile = BytesIO(oosource)
        ootemplatefile = BytesIO(oosource)
        pooutputfile = BytesIO()

        self.conversion_module.convertoo(
            ooinputfile, pooutputfile, ootemplatefile, targetlanguage="en-US"
        )
        posource = pooutputfile.getvalue()

        poinputfile = BytesIO(posource)
        ootemplatefile = BytesIO(oosource)
        oooutputfile = BytesIO()
        po2oo.convertoo(
            poinputfile, oooutputfile, ootemplatefile, targetlanguage="en-US"
        )
        ooresult = oooutputfile.getvalue().decode("utf-8")
        print(
            "original oo:\n",
            oosource,
            "po version:\n",
            posource,
            "output oo:\n",
            ooresult,
        )
        return ooresult.split("\t")[10]

    def check_roundtrip(self, filename, text):
        """Checks that the text converted to po and back is the same as the original."""
        assert self.roundtripstring(filename, text) == text

    def test_simpleentity(self):
        """checks that a simple oo entry converts properly to a po entry"""
        oosource = r"svx	source\dialog\numpages.src	0	string	RID_SVXPAGE_NUM_OPTIONS	STR_BULLET			0	en-US	Character				20050924 09:13:58"
        pofile = self.convert(oosource)
        pounit = self.singleelement(pofile)
        assert pounit.source == "Character"
        assert pounit.target == ""

    def test_escapes(self):
        """checks that a simple oo entry converts escapes properly to a po entry"""
        oosource = r"wizards	source\formwizard\dbwizres.src	0	string	RID_DB_FORM_WIZARD_START + 19				0	en-US	Newline \n Newline Tab \t Tab CR \r CR				20050924 09:13:58"
        pofile = self.convert(oosource)
        pounit = self.singleelement(pofile)
        poelementsrc = str(pounit)
        print(poelementsrc)
        assert "Newline \n Newline" in pounit.source
        assert "Tab \t Tab" in pounit.source
        assert "CR \r CR" in pounit.source

    def test_roundtrip_escape(self):
        self.check_roundtrip(
            "strings.src",
            r"The given command is not a SELECT statement.\nOnly queries are allowed.",
        )
        self.check_roundtrip(
            "source\\ui\\dlg\\AutoControls_tmpl.hrc",
            r";\t59\t,\t44\t:\t58\t{Tab}\t9\t{Space}\t32",
        )
        self.check_roundtrip(
            "inc_openoffice\\windows\\msi_languages\\Nsis.ulf",
            r"The installation files must be unpacked and copied to your hard disk in preparation for the installation. After that, the %PRODUCTNAME installation will start automatically.\r\n\r\nClick \'Next\' to continue.",
        )
        self.check_roundtrip("file.xhp", r"\<ahelp\>")
        self.check_roundtrip("file.xhp", r"\<ahelp prop=\"value\"\>")
        self.check_roundtrip(
            "file.xhp", r"\<ahelp prop=\"value\"\>marked up text\</ahelp\>"
        )
        self.check_roundtrip("file.xhp", r"\<ahelp prop=\"value>>\"\>")
        self.check_roundtrip(
            "file.xhp",
            r"""\<ahelp prop=\"value>>\"\>'Next'>> or "<<Previous"\</ahelp\>""",
        )
        self.check_roundtrip(
            "address_auto.xhp",
            r"""example, \<item type=\"literal\"\>'Harry\\'s Bar'.\</item\>""",
        )

    def test_roundtrip_whitespaceonly(self):
        """check items that are only special instances of whitespce"""
        self.check_roundtrip("choose_chart_type.xhp", r" ")
        self.check_roundtrip("choose_chart_type.xhp", b"\xc2\xa0".decode("utf-8"))

    def test_double_escapes(self):
        oosource = r"helpcontent2	source\text\shared\01\02100001.xhp	0	help	par_id3150670 35				0	en-US	\\<				2002-02-02 02:02:02"
        pofile = self.convert(oosource)
        pounit = self.singleelement(pofile)
        poelementsrc = str(pounit)
        print(poelementsrc)
        assert pounit.source == r"\<"

    def test_escapes_helpcontent2(self):
        """checks that a helpcontent2 entry converts escapes properly to a po entry"""
        oosource = r"helpcontent2	source\text\smath\guide\parentheses.xhp	0	help	par_id3150344	4			0	en-US	size *2 \\langle x \\rangle				2002-02-02 02:02:02"
        pofile = self.convert(oosource)
        pounit = self.singleelement(pofile)
        poelementsrc = str(pounit)
        print(poelementsrc)
        assert pounit.source == r"size *2 \langle x \rangle"

    def test_msgid_bug_error_address(self):
        """tests the we have the correct url for reporting msgid bugs"""
        oosource = r"wizards	source\formwizard\dbwizres.src	0	string	RID_DB_FORM_WIZARD_START + 19				0	en-US	Newline \n Newline Tab \t Tab CR \r CR				20050924 09:13:58"
        pofile = self.convert(oosource)
        assert pofile.units[0].isheader()
        assert pofile.parseheader()["Report-Msgid-Bugs-To"]
        bug_url = parse.urlparse(pofile.parseheader()["Report-Msgid-Bugs-To"])
        print(bug_url)
        assert bug_url[:3] == ("http", "qa.openoffice.org", "/issues/enter_bug.cgi")
        assert parse.parse_qs(bug_url[4], True) == {
            "comment": [""],
            "component": ["l10n"],
            "form_name": ["enter_issue"],
            "short_desc": ["Localization issue in file: "],
            "subcomponent": ["ui"],
        }

    def test_x_comment_inclusion(self):
        """test that we can merge x-comment language entries into comment sections of the PO file"""
        en_USsource = r"wizards	source\formwizard\dbwizres.src	0	string	RID_DB_FORM_WIZARD_START + 19				0	en-US	Text		Quickhelp	Title	20050924 09:13:58"
        xcommentsource = r"wizards	source\formwizard\dbwizres.src	0	string	RID_DB_FORM_WIZARD_START + 19				0	x-comment	%s		%s	%s	20050924 09:13:58"
        # Real comment
        comment = "Comment"
        commentsource = (
            en_USsource + "\n" + xcommentsource % (comment, comment, comment)
        )
        pofile = self.convert(commentsource)
        if isinstance(pofile, poheader):
            units = pofile.units[1:]
        else:
            units = pofile.units
        textunit = units[0]
        assert textunit.source == "Text"
        assert comment in textunit.getnotes("developer")
        quickhelpunit = units[1]
        assert quickhelpunit.source == "Quickhelp"
        assert comment in quickhelpunit.getnotes("developer")
        titleunit = units[2]
        assert titleunit.source == "Title"
        assert comment in titleunit.getnotes("developer")
        # Whitespace and blank
        for comment in ("   ", ""):
            commentsource = (
                en_USsource + "\n" + xcommentsource % (comment, comment, comment)
            )
            pofile = self.convert(commentsource)
            if isinstance(pofile, poheader):
                units = pofile.units[1:]
            else:
                units = pofile.units
            textunit = units[0]
            assert textunit.source == "Text"
            assert textunit.getnotes("developer") == ""
            quickhelpunit = units[1]
            assert quickhelpunit.source == "Quickhelp"
            assert quickhelpunit.getnotes("developer") == ""
            titleunit = units[2]
            assert titleunit.source == "Title"
            assert titleunit.getnotes("developer") == ""


class TestOO2POCommand(test_convert.TestConvertCommand, TestOO2PO):
    """Tests running actual oo2po commands on files"""

    convertmodule = oo2po

    expected_options = [
        "--source-language=LANG",
        "-l LANG, --language=LANG",
        "-P, --pot",
        "--duplicates=DUPLICATESTYLE",
        "--multifile=MULTIFILESTYLE",
        "--nonrecursiveinput",
    ]

    def test_preserve_filename(self):
        """Ensures that the filename is preserved."""
        oosource = rb"svx	source\dialog\numpages.src	0	string	RID_SVXPAGE_NUM_OPTIONS	STR_BULLET			0	en-US	Character				20050924 09:13:58"
        self.create_testfile("snippet.sdf", oosource)
        oofile = oo.oofile(self.open_testfile("snippet.sdf"))
        assert oofile.filename.endswith("snippet.sdf")
        oofile.parse(oosource)
        assert oofile.filename.endswith("snippet.sdf")

    def test_simple_pot(self):
        """tests the simplest possible conversion to a pot file"""
        oosource = r"svx	source\dialog\numpages.src	0	string	RID_SVXPAGE_NUM_OPTIONS	STR_BULLET			0	en-US	Character				20050924 09:13:58"
        self.create_testfile("simple.oo", oosource)
        self.run_command("simple.oo", "simple.pot", pot=True, nonrecursiveinput=True)
        pofile = self.target_filetype(self.open_testfile("simple.pot"))
        poelement = self.singleelement(pofile)
        assert poelement.source == "Character"
        assert poelement.target == ""

    def test_simple_po(self):
        """tests the simplest possible conversion to a po file"""
        oosource1 = r"svx	source\dialog\numpages.src	0	string	RID_SVXPAGE_NUM_OPTIONS	STR_BULLET			0	en-US	Character				20050924 09:13:58"
        oosource2 = r"svx	source\dialog\numpages.src	0	string	RID_SVXPAGE_NUM_OPTIONS	STR_BULLET			0	ku	Karakter				20050924 09:13:58"
        self.create_testfile("simple.oo", oosource1 + "\n" + oosource2)
        self.run_command("simple.oo", "simple.po", lang="ku", nonrecursiveinput=True)
        pofile = self.target_filetype(self.open_testfile("simple.po"))
        poelement = self.singleelement(pofile)
        assert poelement.source == "Character"
        assert poelement.target == "Karakter"

    def test_onefile_nonrecursive(self):
        """tests the --multifile=onefile option and make sure it doesn't produce a directory"""
        oosource = r"svx	source\dialog\numpages.src	0	string	RID_SVXPAGE_NUM_OPTIONS	STR_BULLET			0	en-US	Character				20050924 09:13:58"
        self.create_testfile("simple.oo", oosource)
        self.run_command("simple.oo", "simple.pot", pot=True, multifile="onefile")
        assert os.path.isfile(self.get_testfilename("simple.pot"))

    def test_remove_duplicates(self):
        """test that removing of duplicates works correctly (bug 171)"""
        oosource = r"""
sd	source\ui\animations\SlideTransitionPane.src	0	checkbox	DLG_SLIDE_TRANSITION_PANE	CB_AUTO_PREVIEW	HID_SD_SLIDETRANSITIONPANE_CB_AUTO_PREVIEW		1	en-US	Automatic preview				20060725 03:26:42
sd	source\ui\animations\AnimationSchemesPane.src	0	checkbox	DLG_ANIMATION_SCHEMES_PANE	CB_AUTO_PREVIEW	HID_SD_ANIMATIONSCHEMESPANE_CB_AUTO_PREVIEW		1	en-US	Automatic preview				20060725 03:26:42
sd	source\ui\animations\CustomAnimationCreateDialog.src	0	checkbox	RID_TP_CUSTOMANIMATION_ENTRANCE	CBX_PREVIEW			143	en-US	Automatic preview				20060725 03:26:42
sd	source\ui\animations\CustomAnimationCreateDialog.src	0	checkbox	RID_TP_CUSTOMANIMATION_ENTRANCE	CBX_PREVIEW			143	fr	Aperçu automatique				20060725 03:26:42
sd	source\ui\animations\CustomAnimationSchemesPane.src	0	checkbox	DLG_CUSTOMANIMATION_SCHEMES_PANE	4			0	en-US	Automatic preview				20060725 03:26:42
sd	source\ui\animations\CustomAnimationSchemesPane.src	0	checkbox	DLG_CUSTOMANIMATION_SCHEMES_PANE	4			0	fr	Aperçu automatique				20060725 03:26:42
"""
        self.create_testfile("simple.oo", oosource)
        self.run_command(
            "simple.oo",
            "simple.po",
            language="fr",
            multifile="onefile",
            error="traceback",
            duplicates="merge",
        )
        pofile = self.target_filetype(self.open_testfile("simple.po"))
        assert len(pofile.units) == 2
        assert pofile.units[1].target == "Aperçu automatique"



from translate.lang import data


def test_languagematch():
    """test language comparison"""
    # Simple comparison
    assert data.languagematch("af", "af")
    assert not data.languagematch("af", "en")

    # Handle variants
    assert data.languagematch("pt", "pt_PT")
    # FIXME don't think this one is correct
    # assert not data.languagematch("sr", "sr@Latn")

    # No first language code, we just check that the other code is valid
    assert data.languagematch(None, "en")
    assert data.languagematch(None, "en_GB")
    assert data.languagematch(None, "en_GB@Latn")
    assert not data.languagematch(None, "not-a-lang-code")


def test_normalise_code():
    """test the normalisation of language codes"""
    assert data.normalize_code("af_ZA") == "af-za"
    assert data.normalize_code("xx@Latin") == "xx-latin"


def test_simplify_to_common():
    """test language code simplification"""
    assert data.simplify_to_common("af_ZA") == "af"
    assert data.simplify_to_common("pt_PT") == "pt"
    assert data.simplify_to_common("pt_BR") == "pt_BR"


def test_language_names():
    _ = data.tr_lang("en_US")
    assert _("Bokmål, Norwegian; Norwegian Bokmål") == "Norwegian Bokmål"
    assert _("Spanish; Castillian") == "Spanish"
    assert _("Mapudungun; Mapuche") == "Mapudungun"
    assert (
        _("Interlingua (International Auxiliary Language Association)") == "Interlingua"
    )


def test_language_iso_fullname():
    """Test language ISO fullnames."""
    assert data.get_language_iso_fullname("af") == "Afrikaans"
    assert data.get_language_iso_fullname("cak") == "Kaqchikel"
    assert data.get_language_iso_fullname("en_ZA") == "English (South Africa)"
    assert data.get_language_iso_fullname("pt") == "Portuguese"
    assert data.get_language_iso_fullname("pt_PT") == "Portuguese (Portugal)"
    assert data.get_language_iso_fullname("pt_BR") == "Portuguese (Brazil)"
    assert data.get_language_iso_fullname("pt_br") == "Portuguese (Brazil)"
    assert data.get_language_iso_fullname("pt-BR") == "Portuguese (Brazil)"
    assert data.get_language_iso_fullname("pt-br") == "Portuguese (Brazil)"
    assert data.get_language_iso_fullname("ca") == "Catalan"
    assert data.get_language_iso_fullname("ca@valencia") == ""
    assert data.get_language_iso_fullname("") == ""
    assert data.get_language_iso_fullname("z") == ""
    assert data.get_language_iso_fullname("zzz") == ""
    assert data.get_language_iso_fullname("zzzz") == ""
    assert data.get_language_iso_fullname("zz_BB") == ""
    assert data.get_language_iso_fullname("zz-BB") == ""
    assert data.get_language_iso_fullname("zzz_BBB") == ""
    assert data.get_language_iso_fullname("zzz_BBB") == ""


def test_country_iso_name():
    """Test country ISO names."""
    assert data.get_country_iso_name("ZA") == "South Africa"
    assert data.get_country_iso_name("PT") == "Portugal"
    assert data.get_country_iso_name("BR") == "Brazil"
    assert data.get_country_iso_name("br") == "Brazil"
    assert data.get_country_iso_name("ESP") == "Spain"
    assert data.get_country_iso_name("") == ""
    assert data.get_country_iso_name("z") == ""
    assert data.get_country_iso_name("zzz") == ""
    assert data.get_country_iso_name("zzzz") == ""

    # Use common name if available
    assert data.get_country_iso_name("TW") == "Taiwan"
    assert data.get_country_iso_name("TW") != "Taiwan, Province of China"


def test_language_iso_name():
    """Test language ISO names."""
    assert data.get_language_iso_name("af") == "Afrikaans"
    assert data.get_language_iso_name("afr") == "Afrikaans"
    assert data.get_language_iso_name("cak") == "Kaqchikel"
    assert data.get_language_iso_name("en") == "English"
    assert data.get_language_iso_name("pt") == "Portuguese"
    assert data.get_language_iso_name("") == ""
    assert data.get_language_iso_name("z") == ""
    assert data.get_language_iso_name("zzz") == ""
    assert data.get_language_iso_name("zzzz") == ""

    # Use common name if available
    assert data.get_language_iso_name("bn") == "Bangla"
    assert data.get_language_iso_name("bn") != "Bengali"
